///////////////////////////////////////////////////////////////////////////////
//
// GameContainer.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "GameContainer.h"
#include "GameFactory.h"
#include "GameTypes.hrh"
#include "Game.pan"
#include <s32strm.h>
#include <barsread.h>

const TUid KUidContainer = { KGameObjType0_Container };
const TUid KUidContainerMap = { KGameObjType0_ContainerMap };


///////////////////////////////////////////////////////////////////////////////
// CGameContainerBase
///////////////////////////////////////////////////////////////////////////////
CGameContainerBase::CGameContainerBase(TBool aOwnsContents)
: iOwnsContents(aOwnsContents)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameContainerBase::~CGameContainerBase()
	{
	if(iObjects)
		{
		if(iOwnsContents)
			iObjects->ResetAndDestroy();
		else
			iObjects->Reset();
		delete iObjects;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::ResetL(TInt aSize)
	{
	if(iObjects)
		{
		if(iOwnsContents)
			iObjects->ResetAndDestroy();
		else
			iObjects->Reset();
		}
	else
		{
		iObjects = new(ELeave) CArrayPtrFlat<MGameStreamable>(10);
		}
	if(aSize)
		{
		static_cast<CArrayPtrFlat<MGameStreamable>* >(iObjects)->SetReserveL(aSize);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::InternalizeL(TResourceReader& aReader)
	{
	InternalizeContentsL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::InternalizeL(RReadStream& aStream)
	{
	InternalizeContentsL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::ExternalizeL(RWriteStream& aStream) const
	{
	ExternalizeContentsL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::SetId(TInt32)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::InternalizeContentsL(TResourceReader& aReader)
	{
	ResetL();
	TUidType uidType;
	TInt objCount = aReader.ReadInt16();
	static_cast<CArrayPtrFlat<MGameStreamable>*>(iObjects)->SetReserveL(objCount);
	while(objCount--)
		{
		GameFactory::ReadGameTypeL(aReader, uidType);
		MGameStreamable* obj = NewByTypeLC(uidType);
		obj->InternalizeL(aReader);
		iObjects->AppendL(obj);
		CleanupStack::Pop();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::InternalizeContentsL(RReadStream& aStream)
	{
	ResetL();
	TUidType uidType;
	TInt objCount = aStream.ReadInt16L();
	static_cast<CArrayPtrFlat<MGameStreamable>*>(iObjects)->SetReserveL(objCount);
	while(objCount--)
		{
		GameFactory::ReadGameTypeL(aStream, uidType);
		MGameStreamable* obj = NewByTypeLC(uidType);
		obj->InternalizeL(aStream);
		iObjects->AppendL(obj);
		CleanupStack::Pop();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerBase::ExternalizeContentsL(RWriteStream& aStream) const
	{
	const TInt objCount = Count();
	aStream.WriteInt16L(objCount);
	TUidType type;
	for(TInt id=0; id<objCount; id++)
		{
		const MGameStreamable* obj = (*iObjects)[id];
		type = obj->Type();
		GameFactory::WriteGameTypeL(aStream, type);
		obj->ExternalizeL(aStream);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameContainerBase::Count() const
	{
	return (iObjects) ? (iObjects->Count()) : (0);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainerArrayBase
///////////////////////////////////////////////////////////////////////////////
CGameContainerArrayBase::CGameContainerArrayBase(TBool aOwnsContents)
: CGameContainerBase(aOwnsContents)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameContainerArrayBase::Type() const
	{
	return TUidType(KUidContainer);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameContainerArrayBase::BaseAddL(MGameStreamable* aObject)
	{
	if(!iObjects)
		{
		ResetL();
		}
	const TInt id = Count();
	iObjects->AppendL(aObject);
	return id;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerArrayBase::BaseInsertL(TInt aId, MGameStreamable* aNewObject)
	{
	if(!iObjects)
		{
		ResetL();
		}
	if(aId < iObjects->Count())
		{
		iObjects->InsertL(aId, aNewObject);
		}
	else if(aId == iObjects->Count())
		{
		iObjects->AppendL(aNewObject);
		}
	else
		{
		User::Leave(KErrBadHandle);
		}
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameContainerArrayBase::BaseRemove(TInt aId)
	{
	MGameStreamable* ret = &Object(aId);
	iObjects->Delete(aId);
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerArrayBase::BaseReplaceL(TInt aId, MGameStreamable* aNewObject)
	{
	__ASSERT_ALWAYS(Exists(aId), Panic(EPanicObjectNotFound));
	MGameStreamable*& obj = (*iObjects)[aId];
	if(iOwnsContents)
		{
		delete obj;
		}
	obj = aNewObject;
	}

///////////////////////////////////////////////////////////////////////////////
TBool CGameContainerArrayBase::Exists(TInt aId) const
	{
	TBool b = (aId >= 0) && (aId < Count());
	if(b)
		b = ((*iObjects)[aId] != NULL);
	return b;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable& CGameContainerArrayBase::Object(TInt aId)
	{
	__ASSERT_ALWAYS(Exists(aId), Panic(EPanicObjectNotFound));
	MGameStreamable* r = (*iObjects)[aId];
	__ASSERT_DEBUG(r, Panic(EPanicObjectNotFound));
	return *r;
	}

///////////////////////////////////////////////////////////////////////////////
const MGameStreamable& CGameContainerArrayBase::Object(TInt aId) const
	{
	__ASSERT_ALWAYS(Exists(aId), Panic(EPanicObjectNotFound));
	return *(iObjects->At(aId));
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameContainerMapBase
///////////////////////////////////////////////////////////////////////////////
CGameContainerMapBase::CGameContainerMapBase(TBool aOwnsContents)
: CGameContainerBase(aOwnsContents), iLastInstanceNum(50)
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameContainerMapBase::Type() const
	{
	return TUidType(KUidContainerMap);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::InternalizeL(TResourceReader& aReader)
	{
	ResetL();
	TUidType uidType;
	TInt objCount = aReader.ReadInt16();
	static_cast<CArrayPtrFlat<MGameStreamable>*>(iObjects)->SetReserveL(objCount);
	while(objCount--)
		{
		GameFactory::ReadGameTypeL(aReader, uidType);
		MGameStreamable* obj = NewByTypeLC(uidType);
		obj->InternalizeL(aReader);
		BaseAddL(obj);
		CleanupStack::Pop();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::InternalizeL(RReadStream& aStream)
	{
	CGameContainerBase::InternalizeL(aStream);
	iLastInstanceNum = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::ExternalizeL(RWriteStream& aStream) const
	{
	CGameContainerBase::ExternalizeL(aStream);
	aStream.WriteInt16L(iLastInstanceNum);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::BaseAddL(MGameStreamable* aObject)
	{
	if(!iObjects)
		{
		ResetL();
		}
	const TInt32 key = ClaimKey();
	SetObjectKey(aObject, key);
	const TInt idx = ArrayIndex(key);
	if(idx >= Count())
		iObjects->AppendL(aObject);
	else
		iObjects->At(idx) = aObject;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameContainerMapBase::BaseRemove(TInt32 aKey)
	{
	MGameStreamable* o = BaseLookupL(aKey);
	if(o)
		{
		const TInt idx = ArrayIndex(aKey);
		iObjects->At(idx) = NULL;
		}
	return o;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::BaseReplaceL(TInt32 aKey, MGameStreamable* aObject)
	{
	MGameStreamable* o = BaseLookupL(aKey);
	if(iOwnsContents)
		{
		delete o;
		}
	o = aObject;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameContainerMapBase::SetObjectKey(MGameStreamable* aObject, TInt32 aKey) const
	{
	aObject->SetId(aKey);
	}

///////////////////////////////////////////////////////////////////////////////
TInt32 CGameContainerMapBase::ObjectKey(MGameStreamable* aObject) const
	{
	return TInt32(aObject->Type()[2].iUid);
	}

///////////////////////////////////////////////////////////////////////////////
TInt32 CGameContainerMapBase::ClaimKey()
	{
	++iLastInstanceNum;
	TInt32 key = CalculateNextKey(iLastInstanceNum);
	return key;
	}

///////////////////////////////////////////////////////////////////////////////
TInt32 CGameContainerMapBase::PeekKey() const
	{
	return CalculateNextKey(TInt16(iLastInstanceNum + 1));
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameContainerMapBase::BaseLookupL(TInt32 aKey) const
	{
	if(!iObjects)
		{
		return NULL;
		}
	MGameStreamable* object = NULL;
	const TInt idx = ArrayIndex(aKey);
	if( (idx >= 0) && (idx < Count()) )
		{
		object = iObjects->At(idx);
		if(object)
			{
			const TInt32 actualKey = ObjectKey(object);
			if(InstanceNum(aKey) != InstanceNum(actualKey))
				object = NULL;
			}
		}
	return object;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameContainerMapBase::BaseFirst(TInt32& aKey) const
	{
	TInt idx = 0;
	const TInt count = Count();
	while(idx < count)
		{
		MGameStreamable* object = iObjects->At(idx);
		if(object != NULL)
			{
			aKey = Key(idx, InstanceNum(ObjectKey(object)));
			return object;
			}
		++idx;
		}
	aKey = Key(-1, 0);
	return NULL;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* CGameContainerMapBase::BaseNext(TInt32& aKey) const
	{
	TInt idx = ArrayIndex(aKey);
	++idx; // Next.
	const TInt count = Count();
	while(idx < count)
		{
		MGameStreamable* object = iObjects->At(idx);
		if(object != NULL)
			{
			aKey = Key(idx, InstanceNum(ObjectKey(object)));
			return object;
			}
		++idx;
		}
	aKey = Key(-1, 0);
	return NULL;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameContainerMapBase::FindFirstAvailableArrayIndex() const
	{
	TInt idx = 0;
	const TInt count = Count();
	while(idx < count)
		{
		if(iObjects->At(idx) == NULL)
			break;
		++idx;
		}
	return idx;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameContainerMapBase::FindNextObject(TInt aIdx) const
	{
	++aIdx; // Next
	const TInt count = Count();
	while(aIdx < count)
		{
		if(iObjects->At(aIdx) != NULL)
			break;
		++aIdx;
		}
	return aIdx;
	}

///////////////////////////////////////////////////////////////////////////////
TInt32 CGameContainerMapBase::CalculateNextKey(TInt16 aInstanceNum) const
	{
	const TInt idx = FindFirstAvailableArrayIndex();
	return Key(idx, aInstanceNum);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
