///////////////////////////////////////////////////////////////////////////////
//
// GameContainer.h
//
//
// Container base classes for persisting an array of polymorphic objects.
// These classes are used for holding game objects and sprites.
//
///////////////////////////////////////////////////////////////////////////////
#ifndef GAMECONTAINER_H
#define GAMECONTAINER_H

#include <e32base.h>
class RReadStream;
class RWriteStream;
class TResourceReader;
//

///////////////////////////////////////////////////////////////////////////////
// MGameStreamable
///////////////////////////////////////////////////////////////////////////////
// Interface for a polymorphic objects.
// Objects have a unique key for identifying them. This key is stored in the
// last element, (index 2), of the TUidType returned by the object's Type()
// function, and may be set using the SetId() function.
///////////////////////////////////////////////////////////////////////////////
class MGameStreamable
	{
public:
	virtual ~MGameStreamable() {}
	virtual TUidType Type() const = 0; // Object type and unique ID.
	virtual void InternalizeL(TResourceReader&) = 0;
	virtual void InternalizeL(RReadStream&) = 0;
	virtual void ExternalizeL(RWriteStream&) const = 0;
	virtual void SetId(TInt32) = 0;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainerBase
///////////////////////////////////////////////////////////////////////////////
// Base class for a polymorphic object container using a dynamic array of
// pointers.
///////////////////////////////////////////////////////////////////////////////
class CGameContainerBase : public CBase, public MGameStreamable
	{
public:
	CGameContainerBase(TBool aOwnsContents);
	virtual ~CGameContainerBase();
public:
	virtual void InternalizeL(TResourceReader&);
	virtual void InternalizeL(RReadStream&);
	virtual void ExternalizeL(RWriteStream&) const;
	virtual void SetId(TInt32);
	TInt Count() const;
	virtual void ResetL(TInt aSize=0);       // Call to initialize.
protected:
	void InternalizeContentsL(TResourceReader&);
	void InternalizeContentsL(RReadStream&);
	void ExternalizeContentsL(RWriteStream&) const;
	virtual MGameStreamable* NewByTypeLC(const TUidType&) = 0;
protected:
	TBool iOwnsContents;
	CArrayPtr<MGameStreamable>* iObjects;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainerArrayBase
///////////////////////////////////////////////////////////////////////////////
// Base class for an array of polymorphic MGameStreamable derived classes.
///////////////////////////////////////////////////////////////////////////////
class CGameContainerArrayBase : public CGameContainerBase
	{
public:
	CGameContainerArrayBase(TBool aOwnsContents);
	virtual TUidType Type() const;
protected:
	TBool Exists(TInt) const;
	MGameStreamable& Object(TInt);
	const MGameStreamable& Object(TInt) const;
protected:
	virtual TInt BaseAddL(MGameStreamable*);
	virtual void BaseInsertL(TInt, MGameStreamable*);
	virtual MGameStreamable* BaseRemove(TInt);
	virtual void BaseReplaceL(TInt, MGameStreamable*);
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainer
///////////////////////////////////////////////////////////////////////////////
// Typed object container. 
// OBJECT_T must be an MGameStreamable derived class. The NewByTypeLC() virtual
// method must be implemented by a derived class.
///////////////////////////////////////////////////////////////////////////////
template<typename OBJECT_T>
class CGameContainer : public CGameContainerArrayBase
	{
public:
	inline CGameContainer(TBool aOwnsContents=ETrue)     : CGameContainerArrayBase(aOwnsContents) { }
	inline OBJECT_T& operator[](TInt aIndex)             { return static_cast<OBJECT_T&>(CGameContainerArrayBase::Object(aIndex)); }
	inline const OBJECT_T& operator[](TInt aIndex) const { return static_cast<const OBJECT_T&>(CGameContainerArrayBase::Object(aIndex)); }
	inline TInt AddL(OBJECT_T* aObject)                  { return CGameContainerArrayBase::BaseAddL(aObject); }
	inline void InsertL(TInt aIndex, OBJECT_T* aObject)  { CGameContainerArrayBase::BaseInsertL(aIndex, aObject); }
	inline OBJECT_T* Remove(TInt aIndex)                 { return static_cast<OBJECT_T*>(CGameContainerArrayBase::BaseRemove(aIndex)); }
	inline void ReplaceL(TInt aIndex, OBJECT_T* aObject) { CGameContainerArrayBase::BaseReplaceL(aIndex, aObject); }
	};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainerMapBase
///////////////////////////////////////////////////////////////////////////////
// Object map base class.
// Permits retrieval of objects by unique ID generated by the map.
// Use of a unique ID to access objects decouples the object data from its
// users, enabling the container to dynamically alter its contents based upon
// current demand.
///////////////////////////////////////////////////////////////////////////////
class CGameContainerMapBase : public CGameContainerBase
	{
public:
	CGameContainerMapBase(TBool aOwnsContents);
	TInt32 ClaimKey();
	TInt32 PeekKey() const;
public:
	virtual TUidType Type() const;
	virtual void InternalizeL(TResourceReader&);
	virtual void InternalizeL(RReadStream&);
	virtual void ExternalizeL(RWriteStream&) const;
protected:
	virtual void BaseAddL(MGameStreamable*);
	virtual MGameStreamable* BaseRemove(TInt32 aKey);
	virtual void BaseReplaceL(TInt32 aKey, MGameStreamable*);
protected:
	virtual void SetObjectKey(MGameStreamable*, TInt32 aKey) const;
	virtual TInt32 ObjectKey(MGameStreamable*) const;
	MGameStreamable* BaseLookupL(TInt32 aKey) const;
	MGameStreamable* BaseFirst(TInt32& aKey) const;
	MGameStreamable* BaseNext(TInt32& aKey) const;
protected:
	TInt FindFirstAvailableArrayIndex() const;
	TInt FindNextObject(TInt aIdx) const;
	TInt32 CalculateNextKey(TInt16 aInstanceNum) const;
	inline TInt32 Key(TInt aArrayIndex, TInt16 aInstanceNum) const;
	inline TInt ArrayIndex(TInt32 aKey) const;
	inline TInt16 InstanceNum(TInt32 aKey) const;
protected:
	TInt16 iLastInstanceNum;
	};

///////////////////////////////////////////////////////////////////////////////
inline TInt32 CGameContainerMapBase::Key(TInt aArrayIndex, TInt16 aInstanceNum) const
	{ return TInt32( (TInt32(aInstanceNum) << 16) | TInt16(aArrayIndex) ); }
inline TInt CGameContainerMapBase::ArrayIndex(TInt32 aKey) const
	{ return TInt(TInt16(aKey)); }
inline TInt16 CGameContainerMapBase::InstanceNum(TInt32 aKey) const
	{ return TInt16(aKey >> 16); }

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameContainerMap
///////////////////////////////////////////////////////////////////////////////
// Typed object map container.
// OBJECT_T must be an MGameStreamable derived class. The NewByTypeLC() virtual
// method must be implemented by a derived class.
// 
///////////////////////////////////////////////////////////////////////////////
template<typename OBJECT_T>
class CGameContainerMap : public CGameContainerMapBase
	{
public:
	inline CGameContainerMap(TBool aOwnsContents=ETrue)  : CGameContainerMapBase(aOwnsContents) { }
	inline void AddL(OBJECT_T* aObject)                  { CGameContainerMapBase::BaseAddL(aObject); }
	inline OBJECT_T* Remove(TInt32 aKey)                 { return static_cast<OBJECT_T*>(CGameContainerMapBase::BaseRemove(aKey)); }
	inline void ReplaceL(TInt32 aKey, OBJECT_T* aObject) { CGameContainerMapBase::BaseReplaceL(aKey, aObject); }
	inline OBJECT_T* LookupL(TInt32 aKey) const          { return static_cast<OBJECT_T*>(CGameContainerMapBase::BaseLookupL(aKey)); }
	inline OBJECT_T* First(TInt32& aKey) const           { return static_cast<OBJECT_T*>(CGameContainerMapBase::BaseFirst(aKey)); }
	inline OBJECT_T* Next(TInt32& aKey) const            { return static_cast<OBJECT_T*>(CGameContainerMapBase::BaseNext(aKey)); }
	};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif

