///////////////////////////////////////////////////////////////////////////////
//
// GameFactory.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "GameFactory.h"
#include "SpriteBitmap.h"
#include "SpriteParticle.h"
#include "SpritePolygon.h"
#include "Objects.h"
#include "State.h"
#include "ObjectStates.h"
#include "PlayerStates.h"
#include "AsteroidState.h"
#include "GameTypes.hrh"
#include <barsread.h>
#include <coemain.h>


///////////////////////////////////////////////////////////////////////////////
// GameFactory
///////////////////////////////////////////////////////////////////////////////
void GameFactory::ReadGameTypeL(TResourceReader& aReader, TUidType& aType)
	{ // static
	TUid type[KMaxCheckedUid];
	for(TInt i=0; i<KMaxCheckedUid; i++)
		{
		type[i].iUid = aReader.ReadInt32();
		}
	aType = TUidType(type[0], type[1], type[2]);
	}

///////////////////////////////////////////////////////////////////////////////
void GameFactory::ReadGameTypeL(RReadStream& aStream, TUidType& aType)
	{ // static
	TUid type[KMaxCheckedUid];
	for(TInt i=0; i<KMaxCheckedUid; i++)
		{
		type[i].iUid = aStream.ReadInt32L();
		}
	aType = TUidType(type[0], type[1], type[2]);
	}

///////////////////////////////////////////////////////////////////////////////
void GameFactory::WriteGameTypeL(RWriteStream& aStream, const TUidType& aType)
	{ // static
	for(TInt i=0; i<KMaxCheckedUid; i++)
		{
		aStream.WriteInt32L(aType[i].iUid);
		}
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* GameFactory::NewByTypeLC(const TUidType& aType, CGameObjectController& aController)
	{ // static
	MGameStreamable* ret = NULL;
	switch(aType[EGameTypeIndexCategory].iUid)
		{
	case KGameObjType0_Sprite:
		ret = NewSpriteByTypeLC(aType);
		break;
	case KGameObjType0_Object:
		ret = NewObjectByTypeLC(aType, aController);
		break;
	case KGameObjType0_ObjectState:
		ret = NewObjectStateByTypeLC(aType, aController);
		break;

	case KGameObjType0_Container:
	case KGameObjType0_ContainerMap:
	default:
		break;
		}
	if(!ret)
		{
#ifdef _DEBUG
		_LIT(KBadClassType,"Bad type 0x%X");
		TBuf<0x100> context;
		context.Format(KBadClassType, aType[EGameTypeIndexCategory].iUid);
		User::Panic(context, aType[EGameTypeIndexCategory].iUid);
#else
		User::Leave(KErrCorrupt);
#endif
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* GameFactory::NewSpriteByTypeLC(const TUidType& aType)
	{ // static
	CCoeEnv* coeEnv = CCoeEnv::Static();
	MGameStreamable* ret = NULL;
	if(aType[EGameTypeIndexCategory].iUid == KGameObjType0_Sprite)
		{
		switch(aType[EGameTypeIndexGroup].iUid)
			{
		case KGameObjType1_SpriteBitmap:
			ret = new(ELeave) CGameBitmapSprite(coeEnv->WsSession(), *(coeEnv->ScreenDevice()));
			CleanupStack::PushL((CGameBitmapSprite*)ret);
			break;

		case KGameObjType1_SpriteParticle:
			ret = new(ELeave) CGameParticleSprite(KRgbWhite);
			CleanupStack::PushL((CGameParticleSprite*)ret);
			break;

		case KGameObjType1_SpritePolygon:
			ret = new(ELeave) CGamePolygonSprite;
			CleanupStack::PushL((CGamePolygonSprite*)ret);
			break;

		default:
			break;
			}
		}
	if(!ret)
		{
#ifdef _DEBUG
		_LIT(KBadSpriteType,"Bad sprite 0x%X");
		TBuf<0x100> context;
		context.Format(KBadSpriteType, aType[EGameTypeIndexGroup].iUid);
		User::Panic(context, aType[EGameTypeIndexGroup].iUid);
#else
		User::Leave(KErrCorrupt);
#endif
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* GameFactory::NewObjectByTypeLC(const TUidType& aType, CGameObjectController& aController)
	{ // static
	MGameStreamable* ret = NULL;
	if(aType[EGameTypeIndexCategory].iUid == KGameObjType0_Object)
		{
		switch(aType[EGameTypeIndexGroup].iUid)
			{
		case KGameObjType1_Object:
			ret = new(ELeave) CGameObject(aController);
			CleanupStack::PushL((CGameObject*)ret);
			break;
		default:
			break;
			}
		}
	if(!ret)
		{
#ifdef _DEBUG
		_LIT(KBadObjectType,"Bad object 0x%X");
		TBuf<0x100> context;
		context.Format(KBadObjectType, aType[EGameTypeIndexGroup].iUid);
		User::Panic(context, aType[EGameTypeIndexGroup].iUid);
#else
		User::Leave(KErrCorrupt);
#endif
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
MGameStreamable* GameFactory::NewObjectStateByTypeLC(const TUidType& aType, CGameObjectController& aController)
	{ // static
	MGameStreamable* ret = NULL;
	if(aType[EGameTypeIndexCategory].iUid == KGameObjType0_ObjectState)
		{
		switch(aType[EGameTypeIndexGroup].iUid)
			{
		case KGameObjType1_ObjectDeadState:
			ret = new(ELeave) CGameObjectDeadState(aController);
			CleanupStack::PushL((CGameObjectDeadState*)ret);
			break;
		case KGameObjType1_ObjectActiveState:
			ret = new(ELeave) CGameObjectActiveState(aController);
			CleanupStack::PushL((CGameObjectActiveState*)ret);
			break;
		case KGameObjType1_ObjectMotiveState:
			ret = new(ELeave) CGameObjectMotiveState(aController);
			CleanupStack::PushL((CGameObjectMotiveState*)ret);
			break;
		case KGameObjType1_ObjectMotiveSolidState:
			ret = new(ELeave) CGameObjectMotiveSolidState(aController);
			CleanupStack::PushL((CGameObjectMotiveSolidState*)ret);
			break;
		case KGameObjType1_ObjectPlayerState:
			ret = new(ELeave) CGameObjectPlayerState(aController);
			CleanupStack::PushL((CGameObjectPlayerState*)ret);
			break;
		case KGameObjType1_ObjectAsteroidState:
			ret = new(ELeave) CGameObjectAsteroidState(aController);
			CleanupStack::PushL((CGameObjectAsteroidState*)ret);
			break;
		case KGameObjType1_ObjectExplosionState:
			ret = new(ELeave) CGameObjectExplosionState(aController);
			CleanupStack::PushL((CGameObjectExplosionState*)ret);
			break;
		case KGameObjType1_ObjectPickupState:
			ret = new(ELeave) CGameObjectPlayerPickupState(aController);
			CleanupStack::PushL((CGameObjectPlayerPickupState*)ret);
			break;
		default:
			break;
			}
		}
	if(!ret)
		{
#ifdef _DEBUG
		_LIT(KBadStateType,"Bad state 0x%X");
		TBuf<0x100> context;
		context.Format(KBadStateType, aType[EGameTypeIndexGroup].iUid);
		User::Panic(context, aType[EGameTypeIndexGroup].iUid);
#else
		User::Leave(KErrCorrupt);
#endif
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
