///////////////////////////////////////////////////////////////////////////////
//
// ScreenFurniture.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ScreenFurniture.h"
#include "ScreenUtility.h"
#include "GameResourceUtils.h"
#include "Trigonometry.h"
#include <barsread.h>
#include <s32strm.h>
#include "Game.pan"

#define ASSERT_CONSTRUCTED __ASSERT_DEBUG(iBitmaps && iMasks, Panic(EScreenBackgroundNotConstructed))

///////////////////////////////////////////////////////////////////////////////
// CScreenBackground
///////////////////////////////////////////////////////////////////////////////
CScreenBackground::CScreenBackground()
: iBaseColor(KRgbWhite)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CScreenBackground::~CScreenBackground()
	{
	Reset();
	delete iBitmaps;
	delete iMasks;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::ConstructL()
	{
	if(!iBitmaps)
		{
		iBitmaps = new(ELeave) CArrayFixFlat<TBitmap>(2);
		}
	if(!iMasks)
		{
		iMasks = new(ELeave) CArrayPtrFlat<CFbsBitmap>(2);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::InternalizeL(TResourceReader& aReader)
	{
	Reset();
	ConstructL();
	iSize = GameResourceUtils::ReadTSize(aReader);
	iBaseColor.SetRed(aReader.ReadInt8());
	iBaseColor.SetGreen(aReader.ReadInt8());
	iBaseColor.SetBlue(aReader.ReadInt8());
	TPtrC mbmFile(aReader.ReadTPtrC());
	TPoint pos;
	CFbsBitmap* bitmap = NULL;
	TInt c = aReader.ReadInt16();
	while(c--)
		{
		pos = GameResourceUtils::ReadTPoint(aReader);
		bitmap = new(ELeave) CFbsBitmap;
		CleanupStack::PushL(bitmap);
		User::LeaveIfError( bitmap->Load(mbmFile, aReader.ReadInt16()) );
		const TInt mask = aReader.ReadInt16();
		AddBitmapL(pos, bitmap, mask);
		CleanupStack::Pop(bitmap);
		}
	mbmFile.Set(aReader.ReadTPtrC());
	c = aReader.ReadInt16();
	while(c--)
		{
		bitmap = new(ELeave) CFbsBitmap;
		CleanupStack::PushL(bitmap);
		User::LeaveIfError( bitmap->Load(mbmFile, aReader.ReadInt16()) );
		AddMaskL(bitmap);
		CleanupStack::Pop(bitmap);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::InternalizeL(RReadStream& aStream)
	{
	Reset();
	ConstructL();
	aStream >> iSize;
	aStream >> iBaseColor;
	TPoint pos;
	CFbsBitmap* bitmap = NULL;
	TInt c = (TInt)aStream.ReadUint8L();
	while(c--)
		{
		aStream >> pos;
		bitmap = new(ELeave) CFbsBitmap;
		CleanupStack::PushL(bitmap);
		aStream >> *bitmap;
		const TInt mask = aStream.ReadInt16L();
		AddBitmapL(pos, bitmap, mask);
		CleanupStack::Pop(bitmap);
		}
	c = (TInt)aStream.ReadUint8L();
	while(c--)
		{
		bitmap = new(ELeave) CFbsBitmap;
		CleanupStack::PushL(bitmap);
		aStream >> *bitmap;
		AddMaskL(bitmap);
		CleanupStack::Pop(bitmap);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::ExternalizeL(RWriteStream& aStream) const
	{
	ASSERT_CONSTRUCTED;
	aStream << iSize;
	aStream << iBaseColor;
	TInt total = iBitmaps->Count();
	aStream.WriteUint8L((TUint)total);
	TInt c = 0;
	for(; c<total; c++)
		{
		TBitmap& bitmap = iBitmaps->At(c);
		aStream << bitmap.iPos;
		aStream << *(bitmap.iBitmap);
		aStream.WriteInt16L(bitmap.iMaskIndex);
		}
	total = iMasks->Count();
	aStream.WriteUint8L((TUint)total);
	for(c=0; c<total; c++)
		{
		aStream << *(iMasks->At(c));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::Draw(CBitmapContext& aGc)
	{
	ASSERT_CONSTRUCTED;
	TRect r(iSize);
	aGc.Reset();
	aGc.SetPenStyle(CGraphicsContext::ENullPen);
	aGc.SetBrushStyle(CGraphicsContext::ESolidBrush);
	aGc.SetBrushColor(iBaseColor);
	aGc.DrawRect(r);

	const TInt bitmapCount = iBitmaps->Count();
	if(!bitmapCount)
		return;
	const TInt maskCount = iMasks->Count();
	TBitmapDrawer bitmapDrawer;
	bitmapDrawer.SetAlignment(EHCenterVCenter);
	CFbsBitmap* mask = NULL;
	for(TInt c=0; c<bitmapCount; c++)
		{
		TBitmap& bitmap = iBitmaps->At(c);
		r.SetRect(bitmap.iPos, bitmap.iBitmap->SizeInPixels());
		mask = ( (bitmap.iMaskIndex >= 0) && (bitmap.iMaskIndex < maskCount) ) ? (iMasks->At(bitmap.iMaskIndex)) : (NULL);
		bitmapDrawer.Set(*(bitmap.iBitmap), mask);
		bitmapDrawer.Draw(aGc, r);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::Reset()
	{
	if(iBitmaps)
		{
		TInt c = iBitmaps->Count();
		while(c--)
			{
			TBitmap& bitmap = iBitmaps->At(c);
			delete bitmap.iBitmap;
			iBitmaps->Delete(c);
			}
		}
	if(iMasks)
		{
		iMasks->ResetAndDestroy();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::SetSize(const TSize& aSize)
	{
	iSize = aSize;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBackground::SetBaseColor(const TRgb& aColor)
	{
	iBaseColor = aColor;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CScreenBackground::AddBitmapL(CFbsBitmap* aBitmap, TInt aMaskIndex)
	{
	TPoint pos(0, 0);
	return AddBitmapL(pos, aBitmap, aMaskIndex);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CScreenBackground::AddBitmapL(const TPoint& aPos, CFbsBitmap* aBitmap, TInt aMaskIndex)
	{
	ASSERT_CONSTRUCTED;
	TBitmap bitmap;
	bitmap.iPos = aPos;
	bitmap.iBitmap = aBitmap;
	bitmap.iMaskIndex = aMaskIndex;
	const TInt idx = iBitmaps->Count();
	iBitmaps->AppendL(bitmap);
	return idx;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CScreenBackground::AddMaskL(CFbsBitmap* aBitmap)
	{
	ASSERT_CONSTRUCTED;
	const TInt idx = iMasks->Count();
	iMasks->AppendL(aBitmap);
	return idx;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CScreenBox
///////////////////////////////////////////////////////////////////////////////
CScreenBox::CScreenBox(CScreenBitmap& aScreen)
: iAlignment(CGraphicsContext::ELeft), iScreen(aScreen)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::ConstructL(const TRect& aRect)
	{
	SetRectL(aRect);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::InternalizeL(TResourceReader& aReader)
	{
	iRect.iTl = GameResourceUtils::ReadTPoint(aReader);
	iRect.iBr = GameResourceUtils::ReadTPoint(aReader);
	iAlignment = CGraphicsContext::TTextAlign(aReader.ReadUint8());
	iFillStyle.InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::InternalizeL(RReadStream& aStream)
	{
	aStream >> iRect;
	iAlignment = CGraphicsContext::TTextAlign(aStream.ReadUint8L());
	iFillStyle.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::ExternalizeL(RWriteStream& aStream) const
	{
	aStream << iRect;
	aStream.WriteUint8L(iAlignment);
	iFillStyle.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::SetRectL(const TRect& aRect)
	{
	iRect = aRect;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::Draw(CGraphicsContext& aGc)
	{
	aGc.Reset();
	aGc.SetPenStyle(CGraphicsContext::ENullPen);
	iFillStyle.Set(aGc);
	aGc.DrawRect(iRect);
	iFillStyle.UnSet(aGc);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::SetAlignment(CGraphicsContext::TTextAlign aAlignment)
	{
	iAlignment = aAlignment;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBox::SetFillStyle(const TScreenFillStyle& aStyle)
	{
	iFillStyle = aStyle;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenBoxList
///////////////////////////////////////////////////////////////////////////////
CScreenBoxList::CScreenBoxList(CScreenBitmap& aScreen)
: CScreenBox(aScreen)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CScreenBoxList::~CScreenBoxList()
	{
	iList.ResetAndDestroy();
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::ConstructL(const TPoint& aPoint)
	{
	iRect.iTl = aPoint;
	iRect.iBr = iRect.iTl + TSize(1, 1);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::SetRectL(const TRect& aRect)
	{
	const TInt oldHeight = iRect.Height();
	const TInt newHeight = aRect.Height();
	TRect r(aRect);
	TInt c = Count();
	TTrigUnit u;
	while(c--)
		{
		CScreenBox& box = *iList[c];
		u = box.Rect().Height();
		u /= oldHeight;
		u *= newHeight;
		r.iTl.iY = Max(aRect.iTl.iY, r.iBr.iY - u.NearestInt());
		// TODO: Align child box.
		box.SetRectL(r);
		r.iBr.iY = r.iTl.iY;
		}
	CScreenBox::SetRectL(aRect);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::Draw(CGraphicsContext& aGc)
	{
	TInt c = Count();
	while(c--)
		{
		iList[c]->Draw(aGc);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::SetPositionL(const TPoint& aPoint)
	{
	TRect r;
	TInt c = Count();
	while(c--)
		{
		CScreenBox& box = *iList[c];
		r.SetRect(aPoint, box.Rect().Size());
		// TODO: Align child box.
		box.SetRectL(r);
		}
	r.SetRect(aPoint, iRect.Size());
	SetRectL(r);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::ResizeL(const TSize& aSize)
	{
	const TRect r(iRect.iTl, aSize);
	SetRectL(r);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::AddBoxL(CScreenBox* aBox, const TSize& aBoxSize, TInt aIndex)
	{
	TRect r(iRect);
	if(aIndex < 0)
		{
		aIndex = Count();
		if(aIndex)
			r = iList[aIndex - 1]->Rect();
		}
	else if(aIndex < Count())
		r = iList[aIndex]->Rect();
	const TInt w = Max(aBoxSize.iWidth, iRect.Width());
	// TODO: Align child box.
	r.iBr.iX = r.iTl.iX + w;
	r.iTl.iY = r.iBr.iY;
	r.iBr.iY = r.iTl.iY + aBoxSize.iHeight;
	aBox->SetRectL(r);
	if(aIndex < Count())
		{
		User::LeaveIfError(iList.Insert(aBox, aIndex));
		}
	else
		{
		User::LeaveIfError(iList.Append(aBox));
		}
	ShiftLines(aIndex + 1, aBoxSize.iHeight, w);
	iRect.iBr.iX = iRect.iTl.iX + w;
	iRect.iBr.iY += aBoxSize.iHeight;
	}

///////////////////////////////////////////////////////////////////////////////
CScreenBox* CScreenBoxList::RemoveBox(TInt aIndex)
	{
	CScreenBox* box = iList[aIndex];
	const TInt boxH = box->Rect().Height();
	iList.Remove(aIndex);
	ShiftLines(aIndex, -boxH, iRect.Width());
	iRect.iBr.iY -= boxH;
	return box;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::Reset()
	{
	iList.ResetAndDestroy();
	iRect.iBr.iY = iRect.iTl.iY;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenBoxList::ShiftLines(TInt aIndex, TInt aShiftBy, TInt aBoxWidth)
	{
	TRect r;
	const TInt num = Count();
	for(; aIndex<num; aIndex++)
		{
		CScreenBox& box = *iList[aIndex];
		r = box.Rect();
		r.SetWidth(aBoxWidth);
		// TODO: Align child box.
		r.iTl.iY += aShiftBy;
		r.iBr.iY += aShiftBy;
		box.SetRectL(r);
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenTextBox
///////////////////////////////////////////////////////////////////////////////
CScreenTextBox::CScreenTextBox(CScreenBitmap& aScreen)
: CScreenBox(aScreen),
  iPenStyle(CGraphicsContext::ESolidPen, KRgbBlack)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CScreenTextBox::~CScreenTextBox()
	{
	delete iText;
	ReleaseFont();
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::ConstructL(const TRect& aRect, const TFontSpec& aFont)
	{
	iRect = aRect;
	SetFontL(aFont);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::InternalizeL(TResourceReader& aReader)
	{
	CScreenBox::InternalizeL(aReader);
	iPenStyle.InternalizeL(aReader);
	SetRectL(iRect);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::InternalizeL(RReadStream& aStream)
	{
	CScreenBox::InternalizeL(aStream);
	iPenStyle.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::ExternalizeL(RWriteStream& aStream) const
	{
	CScreenBox::ExternalizeL(aStream);
	iPenStyle.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::SetRectL(const TRect& aRect)
	{
	iRect = aRect;
	SetFontL(iFontSpec);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::SetFontL(const TFontSpec& aFont)
	{
	iFontSpec = aFont;
	iFontSpec.iHeight = FontDevice().VerticalPixelsToTwips(iRect.Height() - 2);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::SetTextL(const TDesC& aText)
	{
	HBufC* txt = aText.AllocL();
	delete iText;
	iText = txt;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::Draw(CGraphicsContext& aGc)
	{
	TPtrC txt;
	if(iText)
		{
		txt.Set(*iText);
		}
	DrawText(aGc, txt, iRect);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::SetPenStyle(const TScreenPenStyle& aStyle)
	{
	iPenStyle = aStyle;
	iPenStyle.iFont = iFont;
	}

#include <eikenv.h>
///////////////////////////////////////////////////////////////////////////////
MGraphicsDeviceMap& CScreenTextBox::FontDevice()
	{
//	return *iScreen.Device();
	return *(CEikonEnv::Static()->ScreenDevice());
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::LoadFont()
	{
	CFont* font = NULL;
	const TInt err = FontDevice().GetNearestFontInTwips(font, iFontSpec);
	if(err == KErrNone)
		{
		ReleaseFont();
		iFont = font;
		iPenStyle.iFont = iFont;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::ReleaseFont()
	{
	if(iFont)
		{
		FontDevice().ReleaseFont(iFont);
		iFont = NULL;
		iPenStyle.iFont = NULL;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenTextBox::DrawText(CGraphicsContext& aGc, const TDesC& aText, const TRect& aRect)
	{
	aGc.Reset();
	LoadFont();
	if(iFont)
		{
		iPenStyle.Set(aGc);
		iFillStyle.Set(aGc);
		const TInt offset = aRect.Height() - (iFont->HeightInPixels() - iFont->AscentInPixels());
		aGc.DrawText(aText, aRect, offset, iAlignment);
		iFillStyle.UnSet(aGc);
		iPenStyle.UnSet(aGc);
		ReleaseFont();
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenNumberBox
///////////////////////////////////////////////////////////////////////////////
CScreenNumberBox::CScreenNumberBox(CScreenBitmap& aScreen)
: CScreenTextBox(aScreen)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenNumberBox::SetNumberL(TInt aVal)
	{
	TBuf<16> txt;
	_LIT(KFmt,"%d");
	txt.Format(KFmt, aVal);
	SetTextL(txt);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenNumberBox::SetNumberL(TUint aVal)
	{
	TBuf<16> txt;
	_LIT(KFmt,"%u");
	txt.Format(KFmt, aVal);
	SetTextL(txt);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenNumberBox::SetNumberL(const TTrigUnit& aVal)
	{
	TBuf<64> txt;
	aVal.Format(txt);
	SetTextL(txt);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenMultilineTextBox
///////////////////////////////////////////////////////////////////////////////
CScreenMultilineTextBox::CScreenMultilineTextBox(CScreenBitmap& aScreen)
: CScreenTextBox(aScreen), iNumLines(1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::ConstructL(const TRect& aRect, const TFontSpec& aFont, TInt aNumLines)
	{
	iNumLines = Max(1, aNumLines);
	CScreenTextBox::ConstructL(aRect, aFont);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::InternalizeL(TResourceReader& aReader)
	{
	iNumLines = (TInt)aReader.ReadUint8();
	CScreenTextBox::InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::InternalizeL(RReadStream& aStream)
	{
	iNumLines = (TInt)aStream.ReadUint8L();
	CScreenTextBox::InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteUint8L(TUint(iNumLines));
	CScreenTextBox::ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::SetNumLinesL(TInt aLines, TBool aResize)
	{
	aLines = Max(1, aLines);
	TRect r(iRect);
	if(aResize)
		{
		r.iBr.iY = r.iTl.iY + (LineHeight() * (aLines - iNumLines));
		}
	iNumLines = aLines;
	SetRectL(r);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::SetLineHeightL(TInt aHeight)
	{
	TRect r(iRect);
	r.iBr.iY = r.iTl.iY + (aHeight * iNumLines);
	SetRectL(r);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CScreenMultilineTextBox::LineHeight() const
	{
	return ((iRect.Height() - (iNumLines + 1)) / iNumLines) + 1;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::SetFontL(const TFontSpec& aFont)
	{
	iFontSpec = aFont;
	iFontSpec.iHeight = FontDevice().VerticalPixelsToTwips(LineHeight() - 1);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenMultilineTextBox::Draw(CGraphicsContext& aGc)
	{
	TPtrC txt;
	TRect r(iRect);
	if(iText)
		{
		r.iBr.iY = r.iTl.iY + LineHeight();
		txt.Set(*iText);
		TInt i = txt.Locate(TChar('\n')) + 1;
		while(i > 0)
			{
			CScreenTextBox::DrawText(aGc, txt.Left(i), r);
			txt.Set(txt.Right(txt.Length() - i));
			i = txt.Locate(TChar('\n')) + 1;
			r.iTl.iY += LineHeight();
			r.iBr.iY += LineHeight();
			}
		}
	CScreenTextBox::DrawText(aGc, txt, r);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenScrollingTextBox
///////////////////////////////////////////////////////////////////////////////
CScreenScrollingTextBox::CScreenScrollingTextBox(CScreenBitmap& aScreen)
: CScreenTextBox(aScreen), iScrollRate(1), iScroll(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::ConstructL(const TRect& aRect, const TFontSpec& aFont, TInt aScrollRate)
	{
	CScreenTextBox::ConstructL(aRect, aFont);
	SetScrollRate(aScrollRate);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::InternalizeL(TResourceReader& aReader)
	{
	CScreenTextBox::InternalizeL(aReader);
	iScrollRate = (TInt)aReader.ReadUint8();
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::InternalizeL(RReadStream& aStream)
	{
	CScreenTextBox::InternalizeL(aStream);
	iScrollRate = (TInt)aStream.ReadUint8L();
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::ExternalizeL(RWriteStream& aStream) const
	{
	CScreenTextBox::ExternalizeL(aStream);
	aStream.WriteUint8L(TUint(iScrollRate));
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::SetScrollRate(TInt aRate)
	{
	iScrollRate = aRate;
	iScroll = 0;
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenScrollingTextBox::Draw(CGraphicsContext& aGc)
	{
	TPtrC txt;
	if(iText)
		{
		const TInt l = iText->Length();
		if(l <= 0)
			{
			return;
			}
		const TInt c = l - Font()->TextCount(*iText, iScroll);
		++iScroll;
		if(c > 0)
			{
			txt.Set(iText->Right(c));
			}
		else
			{
			iScroll = 0;
			txt.Set(*iText);
			}
		}
	CScreenTextBox::DrawText(aGc, txt, iRect);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CScreenSpriteBox
///////////////////////////////////////////////////////////////////////////////
CScreenSpriteBox::CScreenSpriteBox(CScreenBitmap& aScreen, CGameSpriteTable& aSpriteTable)
: CScreenBox(aScreen), iSprite(aSpriteTable)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CScreenSpriteBox::~CScreenSpriteBox()
	{
	delete iTextBox;
	iSprite.Close();
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::ConstructL(const TPoint& aPoint, TSpriteId aSpriteId, CScreenTextBox* aText)
	{
	iSprite.OpenL(aSpriteId);
	const TRect r(aPoint, iSprite.FrameSize());
	CScreenBox::ConstructL(r);
	SetTextBox(aText);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::InternalizeL(TResourceReader& aReader)
	{
	CScreenBox::InternalizeL(aReader);
	iSprite.InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::InternalizeL(RReadStream& aStream)
	{
	CScreenBox::InternalizeL(aStream);
	iSprite.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::ExternalizeL(RWriteStream& aStream) const
	{
	CScreenBox::ExternalizeL(aStream);
	iSprite.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::SetRectL(const TRect& aRect)
	{
	CScreenBox::SetRectL(aRect);
	if(iTextBox)
		{
		TRect r(aRect);
		r.iTl.iX += iSprite.FrameSize().iWidth + 1;
		iTextBox->SetRectL(r);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::Draw(CGraphicsContext& aGc)
	{
	aGc.Reset();
	iSprite.DrawFrame(static_cast<CBitmapContext&>(aGc), iRect.iTl);
	if(iTextBox)
		{
		iTextBox->Draw(aGc);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CScreenSpriteBox::SetTextBox(CScreenTextBox* aTextBox)
	{
	if(iTextBox)
		{
		delete iTextBox;
		}
	iTextBox = aTextBox;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

