///////////////////////////////////////////////////////////////////////////////
//
// AsteroidState.cpp
//

///////////////////////////////////////////////////////////////////////////////
#include "AsteroidState.h"
#include "PlayerStates.h"
#include "ObjectController.h"
#include "SpritePolygon.h"
#include "GameResourceUtils.h"
#include "GameTypes.hrh"
#include <s32strm.h>
#include <barsread.h>

const TUid KUid0ObjectState = { KGameObjType0_ObjectState };

///////////////////////////////////////////////////////////////////////////////
// CGameObjectAsteroidState
///////////////////////////////////////////////////////////////////////////////
CGameObjectAsteroidState::CGameObjectAsteroidState(CGameObjectController& aController)
: CGameObjectMotiveSolidState(aController),
	iCollisionMassFactor(0.5),
	iMinMass(1),
	iMaxMass(100)
	{
	iId = KGameObjType1_ObjectAsteroidState;
	}

///////////////////////////////////////////////////////////////////////////////
CGameObjectAsteroidState::~CGameObjectAsteroidState()
	{
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectAsteroidState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectAsteroidState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectAsteroidState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectMotiveState::InternalizeL(aReader);
	iChildren.InternalizeL(aReader);
	iAsteroidColor.InternalizeL(aReader);
	iCollisionMassFactor = GameResourceUtils::ReadTTrigUnit(aReader);
	iMinMass = GameResourceUtils::ReadTTrigUnit(aReader);
	iMaxMass = GameResourceUtils::ReadTTrigUnit(aReader);
	iMinRadius = (TInt)aReader.ReadUint8();
	iMaxRadius = (TInt)aReader.ReadUint8();
	iSplinterMassDeviation = (TInt)aReader.ReadUint8();
	iSplinterMassDeviationRange = (TInt)aReader.ReadUint8();
	iPointValBase = aReader.ReadInt16();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectAsteroidState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectMotiveState::InternalizeL(aStream);
	iChildren.InternalizeL(aStream);
	iAsteroidColor.InternalizeL(aStream);
	iCollisionMassFactor.InternalizeL(aStream);
	iMinMass.InternalizeL(aStream);
	iMaxMass.InternalizeL(aStream);
	iMinRadius = (TInt)aStream.ReadUint8L();
	iMaxRadius = (TInt)aStream.ReadUint8L();
	iSplinterMassDeviation = (TInt)aStream.ReadUint8L();
	iSplinterMassDeviationRange = (TInt)aStream.ReadUint8L();
	iPointValBase = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectAsteroidState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectMotiveState::ExternalizeL(aStream);
	iChildren.ExternalizeL(aStream);
	iAsteroidColor.ExternalizeL(aStream);
	iCollisionMassFactor.ExternalizeL(aStream);
	iMinMass.ExternalizeL(aStream);
	iMaxMass.ExternalizeL(aStream);
	aStream.WriteUint8L(TUint(iMinRadius));
	aStream.WriteUint8L(TUint(iMaxRadius));
	aStream.WriteUint8L(TUint(iSplinterMassDeviation));
	aStream.WriteUint8L(TUint(iSplinterMassDeviationRange));
	aStream.WriteInt16L(iPointValBase);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectAsteroidState::OnEnterL(const TGameMessage& aMessage)
	{
	CGameObjectMotiveState::OnEnterL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	TInt pts = Max(1, o.Mass().NearestInt());
	pts *= iPointValBase;
	o.SetPoints(pts);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectAsteroidState::OnCollisionL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = CGameObjectMotiveState::OnCollisionL(aMessage);
	if(iFlags & EDieOnCollision)
		{
		SendPointsScoreMessageL(aMessage);
		SplinterL(aMessage);

		if(!iController.Random(10))
			{
			CGameObjectState* state = NULL;
			RArray<TInt> pickupStates;
			TInt total = iController.ObjectStates().Count();
			while(total--)
				{
				state = &(iController.ObjectStates()[total]);
				if(state->Type()[1].iUid == KGameObjType1_ObjectPickupState)
					{
					(void)pickupStates.Append(state->Id());
					}
				}
			total = pickupStates.Count();
			if(total)
				{
				total = iController.Random(total);
				total = pickupStates[total];
				pickupStates.Reset();
				state = iController.StateL(total);
				if(state)
					{
					static_cast<CGameObjectPlayerPickupState*>(state)->CreateNewL(ObjectL(aMessage));
					}
				}
			else
				{
				pickupStates.Reset();
				}
			}

		ret = 1;
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectAsteroidState::ChangeObjectSpriteL(const TGameMessage& aMessage)
	{
	if(aMessage.iId == KMsgId_ObjectEnterState)
		{
		CGameObject& o = ObjectL(aMessage);
		CGameSpriteData* pts = ObjectSpritePointsL(o);
		o.SetSpriteData(pts, ETrue);
		return 1;
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectAsteroidState::SplinterL(const TGameMessage& aCollisionMessage)
	{
	CGameObject& o = ObjectL(aCollisionMessage);
	// Number of splinters.
	const TInt numSplinters = iController.Random(3) + 1; // 1 -> 3 splinters
	if(!numSplinters)
		return;
	// Total splinter mass.
	TTrigUnit totalMass(o.Mass());
	totalMass *= iCollisionMassFactor;
	// Average splinter mass.
	TTrigUnit aveMass(totalMass);
	aveMass /= numSplinters;
	if(aveMass < iMinMass)
		return;
	// Impact.
	CGameObject& o2 = ObjectL(aCollisionMessage, CGameObjectState::ESender);
	TVector impact(o2.Velocity());
	impact *= o2.Mass();
	// Find motion of parent after collision.
	TMotionPhysics motion(o.Motion());
	motion.ApplyForce(impact);
	motion.Update();
	// Total momentum.
	TTrigUnit totalMomentum(motion.Velocity().Magnitude());
	totalMomentum *= totalMass;
	// Splinter facing.
	const TDirection facing(motion.Velocity().Direction());
	// Create splinters.
	NewAsteroidsL(numSplinters, o.Position(), TCoOrds(), aveMass, facing, totalMomentum);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectAsteroidState::NewAsteroidsL(TInt aNumAsteroids, const TCoOrds& aPos, const TCoOrds& aPosDelta, const TTrigUnit& aAveMass, TDirection aFacing, const TTrigUnit& aTotalMomentum)
	{
	aFacing -= (aNumAsteroids / 2);
	TMotionPhysics motion;
	TTrigUnit mass;
	TCoOrds pos(aPos);
	for(TInt c=0; c<aNumAsteroids; ++c)
		{
		mass = aAveMass;
		if( (iSplinterMassDeviationRange > 0) && (iSplinterMassDeviation > 0) )
			{
			mass = iController.Random(aAveMass, iSplinterMassDeviationRange, aAveMass / iSplinterMassDeviation);
			}
		if(mass > iMaxMass)
			{
			mass = iMaxMass;
			}
		if(mass >= iMinMass)
			{
			motion.SetPosition(pos);
			motion.SetMass(mass);
			TVector v(aFacing.Vector(aTotalMomentum / motion.Mass()));
			if(Abs(v.iX.Int()) < 1)
				v.iX = TTrigUnit(1);
			motion.SetVelocity(v);
			iChildren.EmitL(iController, motion);
			}
		++aFacing;
		pos += aPosDelta;
		}
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteData* CGameObjectAsteroidState::ObjectSpritePointsL(const CGameObject& aObject)
	{
	CGameSpriteData* pts = new(ELeave) CGameSpriteData;
	CleanupStack::PushL(pts);
	// Colour.
	pts->SetColor(&iAsteroidColor, EFalse);
	// Radius.
	TTrigUnit r(iMaxRadius - iMinRadius);
	r /= (iMaxMass - iMinMass);
	r *= (aObject.Motion().Mass() - iMinMass);
	r += iMinRadius;
	// Centre.
	pts->iTrigXY.iX = r;
	pts->iTrigXY.iY = r;
	pts->iTrigXY /= 2;
	// Vertexes.
	TDirection a;
	TTrigUnit u;
	TTrigXY xy;
	const TTrigUnit rby30(r / 30);
	do
		{
		u = rby30;
		u *= (iController.Random(10) - 10);
		u += r;
		xy = a.Vector(u);
		xy += pts->iTrigXY;
		pts->AppendPointL(TPoint(xy));
		++a;
		}while(a.Value() != 0);

	CleanupStack::Pop(pts);
	return pts;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Global TPoint operators.
///////////////////////////////////////////////////////////////////////////////
TPoint operator*(const TPoint& aPoint, TInt aInt)
	{
	TPoint p(aPoint);
	p.iX *= aInt;
	p.iY *= aInt;
	return p;
	}

///////////////////////////////////////////////////////////////////////////////
TPoint operator/(const TPoint& aPoint, TInt aInt)
	{
	TPoint p(aPoint);
	p.iX /= aInt;
	p.iY /= aInt;
	return p;
	}

///////////////////////////////////////////////////////////////////////////////
TPoint operator+=(TPoint& aPoint, TInt aInt)
	{
	aPoint.iX += aInt;
	aPoint.iY += aInt;
	return aPoint;
	}

///////////////////////////////////////////////////////////////////////////////
TPoint operator-=(TPoint& aPoint, TInt aInt)
	{
	aPoint.iX -= aInt;
	aPoint.iY -= aInt;
	return aPoint;
	}

///////////////////////////////////////////////////////////////////////////////
TPoint operator*=(TPoint& aPoint, TInt aInt)
	{
	aPoint.iX *= aInt;
	aPoint.iY *= aInt;
	return aPoint;
	}

///////////////////////////////////////////////////////////////////////////////
TPoint operator/=(TPoint& aPoint, TInt aInt)
	{
	aPoint.iX /= aInt;
	aPoint.iY /= aInt;
	return aPoint;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

