///////////////////////////////////////////////////////////////////////////////
//
// GameModel.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "GameModel.h"
#include "GameHighScoreTable.h"
#include "Screen.h"
#include "ScreenFurniture.h"
#include "ObjectController.h"
#include "SpriteTable.h"
#include "SpriteBitmap.h"
#include "GameResourceUtils.h"
#include "GameUids.h"
#include "Game.pan"
#include "GameTypes.hrh"
#include <s32stor.h>
#include <barsread.h>
#include <coemain.h>

const TInt KGameTickTimerPriority = CActive::EPriorityStandard;
const TInt32 KGameTickInterval = 1000;  // 1/1000 second

///////////////////////////////////////////////////////////////////////////////
// CGameModel
///////////////////////////////////////////////////////////////////////////////
CGameModel::CGameModel()
: CTimer(KGameTickTimerPriority),
  iGameTick(KGameTickInterval)
	{
	CActiveScheduler::Add(this);
	}

///////////////////////////////////////////////////////////////////////////////
CGameModel::~CGameModel()
	{
	Cancel();
	delete iObjects;
	delete iSprites;
	delete iScreenBackground;
	iScreenBoxes.ResetAndDestroy();
	delete iScreen;
	delete iHighScores;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::ConstructL(RWsSession& aWs, CWsScreenDevice& aScreenDevice, const TSize& aScreenSize, TDisplayMode aScreenMode)
	{
	CTimer::ConstructL();
	iWs = &aWs;
//	iWs->SetAutoFlush(ETrue);
	iScreenDevice = &aScreenDevice;

	CreateScreenL(aScreenSize, aScreenMode);

	iSprites = new(ELeave) CGameSpriteTable;
	iSprites->ResetL();

	iObjects = new(ELeave) CGameObjectController(*iSprites);
	iObjects->ConstructL(NULL, EFalse, NULL, EFalse);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::PlayLevelL(const CStreamStore& aStore, const CStreamDictionary& aDictionary)
	{
	LoadLevelL(aStore, aDictionary);
	PlayL();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadLevelL(const CCoeEnv& aCoeEnv, TInt aResourceId)
	{
	ResetL();

	TResourceReader reader;
	aCoeEnv.CreateResourceReaderLC(reader, aResourceId);
	LoadDetailsL(reader);
	enum TResourceArray { EArraySprites, EArrayObjectStates, EArrayObjects };
	TInt32 resId[EArrayObjects + 1];
	for(TInt c=EArraySprites; c<=EArrayObjects; c++)
		resId[c] = reader.ReadInt32();
	LoadScreenL(reader);
	CleanupStack::PopAndDestroy(); // level resource

	aCoeEnv.CreateResourceReaderLC(reader, resId[EArraySprites]);
	LoadSpritesL(reader);
	CleanupStack::PopAndDestroy(); // array resource
	aCoeEnv.CreateResourceReaderLC(reader, resId[EArrayObjectStates]);
	LoadObjectStatesL(reader);
	CleanupStack::PopAndDestroy(); // array resource
	aCoeEnv.CreateResourceReaderLC(reader, resId[EArrayObjects]);
	LoadObjectsL(reader);
	CleanupStack::PopAndDestroy(); // array resource
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadLevelL(const CStreamStore& aStore, const CStreamDictionary& aDictionary)
	{
	ResetL();

	TStreamId id;
	RStoreReadStream stream;

	id = aDictionary.At(KUidModelDetailsStream);
	stream.OpenLC(aStore, id);
	LoadDetailsL(stream);
	CleanupStack::PopAndDestroy(); // stream

	id = aDictionary.At(KUidModelSpriteStream);
	stream.OpenLC(aStore, id);
	LoadSpritesL(stream);
	CleanupStack::PopAndDestroy(); // stream

	id = aDictionary.At(KUidModelObjectsStream);
	stream.OpenLC(aStore, id);
	LoadObjectsL(stream);
	CleanupStack::PopAndDestroy(); // stream

	id = aDictionary.At(KUidModelScreenStream);
	stream.OpenLC(aStore, id);
	LoadScreenL(stream);
	CleanupStack::PopAndDestroy(); // stream
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SaveLevelL(CStreamStore& aStore, CStreamDictionary& aDictionary, TBool aCommit) const
	{
	TStreamId id;
	RStoreWriteStream stream;

	id = stream.CreateLC(aStore);
	SaveDetailsL(stream);
	stream.CommitL();
	CleanupStack::PopAndDestroy(); // stream
	aDictionary.AssignL(KUidModelDetailsStream, id);

	id = stream.CreateLC(aStore);
	SaveSpritesL(stream);
	stream.CommitL();
	CleanupStack::PopAndDestroy(); // stream
	aDictionary.AssignL(KUidModelSpriteStream, id);

	id = stream.CreateLC(aStore);
	SaveObjectsL(stream);
	stream.CommitL();
	CleanupStack::PopAndDestroy(); // stream
	aDictionary.AssignL(KUidModelObjectsStream, id);

	id = stream.CreateLC(aStore);
	SaveScreenL(stream);
	stream.CommitL();
	CleanupStack::PopAndDestroy(); // stream
	aDictionary.AssignL(KUidModelScreenStream, id);

	if(aCommit)
		{
		aStore.CommitL();
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::ResetL()
	{
	Stop();
	iObjects->Reset();
	iSprites->ResetL();
	iScreenBoxes.ResetAndDestroy();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetScreenSizeL(const TSize& aScreenSize)
	{
	if(aScreenSize != iScreen->Size())
		{
		iScreen->ResizeL(aScreenSize);
		iObjects->SetExtents(aScreenSize);
		}
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameModel::AddScreenBoxL(CScreenBox* aBox)
	{
	const TInt ret = ScreenBoxCount();
	User::LeaveIfError(iScreenBoxes.Append(aBox));
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::DeleteScreenBox(TInt aIndex)
	{
	delete iScreenBoxes[aIndex];
	iScreenBoxes.Remove(aIndex);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameModel::ScreenBoxCount() const
	{
	return iScreenBoxes.Count();
	}

///////////////////////////////////////////////////////////////////////////////
CScreenBox& CGameModel::ScreenBox(TInt aIndex)
	{
	return *iScreenBoxes[aIndex];
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetName(const TDesC& aName)
	{
	iName = aName.Left( Min(KGameMaxLevelNameLen, aName.Length()) );
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetHighScores(CGameHighScoreTable* aHighScores)
	{
	delete iHighScores;
	iHighScores = aHighScores;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::CreateScreenL(const TSize& aScreenSize, TDisplayMode aScreenMode)
	{
	ASSERT(iWs);
	iScreen = new(ELeave) CScreenBitmap;
	iScreen->ConstructL(*iWs, aScreenSize, aScreenMode);

	iScreenBackground = new(ELeave) CScreenBackground;
	iScreenBackground->ConstructL();
	iScreenBackground->SetSize(aScreenSize);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadDetailsL(TResourceReader& aReader)
	{
	iName = aReader.ReadTPtrC();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadDetailsL(RReadStream& aStream)
	{
	TSize size;
	size.iWidth = aStream.ReadInt32L();
	size.iHeight = aStream.ReadInt32L();
	aStream >> iName;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadSpritesL(TResourceReader& aReader)
	{
	iSprites->InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadSpritesL(RReadStream& aStream)
	{
	iSprites->InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadObjectStatesL(TResourceReader& aReader)
	{
	iObjects->InternalizeObjectStatesL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadObjectsL(TResourceReader& aReader)
	{
	iObjects->InternalizeObjectsL(aReader);
	iObjects->SetExtents(iScreen->Size());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadObjectsL(RReadStream& aStream)
	{
	iObjects->InternalizeL(aStream);
	iObjects->SetExtents(iScreen->Size());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadScreenL(TResourceReader& aReader)
	{
	iScreenBackground->InternalizeL(aReader);
	iScreenBackground->SetSize(iScreen->Size());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::LoadScreenL(RReadStream& aStream)
	{
	iScreenBackground->InternalizeL(aStream);
	iScreenBackground->SetSize(iScreen->Size());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SaveDetailsL(RWriteStream& aStream) const
	{
	const TSize size(iScreen->Size());
	aStream.WriteInt32L(size.iWidth);
	aStream.WriteInt32L(size.iHeight);
	aStream << iName;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SaveSpritesL(RWriteStream& aStream) const
	{
	iSprites->ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SaveObjectsL(RWriteStream& aStream) const
	{
	iObjects->ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SaveScreenL(RWriteStream& aStream) const
	{
	iScreenBackground->ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetGameObserver(MGameModelObserver& aObserver)
	{
	iObserver = &aObserver;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetPlayerInput(MGameModelPlayerInput* aPlayerInput)
	{
	iPlayerInput = aPlayerInput;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::PlayL()
	{
	if(iPlayerInput)
		{
		iPlayerInput->ResetPlayerInput(*this);
		}
	NotifyObserverL(MGameModelObserver::EStartGame);
	// Send enter state message.
	const TGameMessage msg(KMsgId_ObjectEnterState, TGameMessage::KNullAddress, TGameMessage::KBroadcastAddress);
	SendMessage(msg);
	QueueGameTick();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::ResumeL()
	{
	NotifyObserverL(MGameModelObserver::EResumeGame);
	QueueGameTick();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::Stop()
	{
	Cancel();
	}

///////////////////////////////////////////////////////////////////////////////
CGameHighScoreTable* CGameModel::HighScores()
	{
	return iHighScores;
	}

///////////////////////////////////////////////////////////////////////////////
CScreenBitmap& CGameModel::Screen()
	{
	return *iScreen;
	}

///////////////////////////////////////////////////////////////////////////////
CGameSpriteTable& CGameModel::Sprites()
	{
	return *iSprites;
	}

///////////////////////////////////////////////////////////////////////////////
CGameObjectController& CGameModel::Objects()
	{
	return *iObjects;
	}

///////////////////////////////////////////////////////////////////////////////
const TDesC& CGameModel::Name() const
	{
	return iName;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameModel::TicksPerSec() const
	{
	return 1000000 / iGameTick;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SetTicksPerSec(TInt aTicks)
	{
	iGameTick = 1000000 / aTicks;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::SendMessage(const TGameMessage& aMessage)
	{
	(void)iObjects->SendMessage(aMessage);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::RunL()
	{
	// Send player input messages.
	if(iPlayerInput)
		{
		TGameMessage message(0, iObjects->MessageAddress(), iPlayerInput->PlayerAddressL(*this));
		TInt8 param = 0;
		iPlayerInput->GetPlayerInput(*this, iPlayerKeyState);
		if(iPlayerKeyState.IsPressed(EPlayerKeyStop))
			{
			message.iId = KMsgId_ObjectStop;
			SendMessage(message);
			}
		else
			{
			if(iPlayerKeyState.IsPressed(EPlayerKeyTurnLeft))
				{
				param = -1;
				message.iId = KMsgId_ObjectRotate;
				message.iParams = TPckgC<TInt8>(param);
				SendMessage(message);
				}
			else if(iPlayerKeyState.IsPressed(EPlayerKeyTurnRight))
				{
				param = 1;
				message.iId = KMsgId_ObjectRotate;
				message.iParams = TPckgC<TInt8>(param);
				SendMessage(message);
				}

			if(iPlayerKeyState.IsPressed(EPlayerKeyThrust))
				{
				message.iId = KMsgId_ObjectThrust;
				SendMessage(message);
				}
			if(iPlayerKeyState.IsPressed(EPlayerKeyShoot))
				{
				message.iId = KMsgId_ObjectShoot;
				SendMessage(message);
				}
			// TODO: For shield. Not tested this...
			if(iPlayerKeyState.IsPressed(EPlayerKeyActivate))
				message.iId = KMsgId_PlayerPowerupActivate;
			else
				message.iId = KMsgId_PlayerPowerupDeActivate;
//			SendMessage(message);
			}
		}
	// Send update message.
	TGameMessage msg(KMsgId_ObjectUpdate, TGameMessage::KNullAddress, TGameMessage::KBroadcastAddress);
	SendMessage(msg);

	// Process collisions.
	msg.iId = KMsgId_ObjectCollision;
	TVector subjectCentre;
	TTrigUnit subjectRadius;
	TInt32 subjectKey = 0;
	CGameObject* subject = iObjects->FirstObject(subjectKey);
	while(subject)
		{
		CGameObjectState* state = iObjects->StateL(subject->StateData().CurrentState());
		if(state)
			{
			// The recipient of any collision messages is 'subject', 'test' is the sender.
			// 'test' itself will be notified when the outer loop gets to it.
			msg.iRecipient = subject->MessageAddress();
			subjectCentre = subject->Centre();
			subjectRadius = subject->Radius();
			CGameObject* test = iObjects->FirstObject(msg.iSender); // Set msg.iSender.
			while(test)
				{
				if(state->TestCollision(*test, subject->Allegience()))
					{
					const TTrigUnit r(test->Radius() + subjectRadius);
					TVector v(test->Centre());
					v -= subjectCentre;
					v.iX = Abs(v.iX);
					v.iY = Abs(v.iY);
					if(r > v.Magnitude()) // Sum of radii > distance between objects.
						state->OnMessageL(msg);   // Send collision message.
					}
				test = iObjects->NextObject(msg.iSender);
				}
			}
		subject = iObjects->NextObject(subjectKey);
		}
	iObjects->ProcessStateChangesAndDiscardDeadObjectsL();

	// Draw.
	CBitmapContext& gc = *(iScreen->Context());
	DrawBackground(gc);
	DrawObjects(gc);
	DrawScreenBoxes(gc);

	// Notify, and queue next tick.
	NotifyObserverL(MGameModelObserver::EGameTick);
	QueueGameTick();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::QueueGameTick()
	{
	After(TTimeIntervalMicroSeconds32(iGameTick));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::NotifyObserverL(MGameModelObserver::TNotification aNotification)
	{
	if(iObserver)
		{
		iObserver->OnGameEventL(*this, aNotification);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::DrawObjects(CBitmapContext& aGc)
	{
	iObjects->Draw(aGc);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::DrawScreenBoxes(CBitmapContext& aGc)
	{
	TInt c = iScreenBoxes.Count();
	while(c--)
		{
		iScreenBoxes[c]->Draw(aGc);
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameModel::DrawBackground(CBitmapContext& aGc)
	{
	if(iScreenBackground)
		{
		iScreenBackground->Draw(aGc);
		}
	else
		{
		aGc.Clear(); // Just blank screen bitmap.
		}
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


