///////////////////////////////////////////////////////////////////////////////
//
// Messages.h
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#ifndef MESSAGES_H
#define MESSAGES_H

#include <e32base.h>
class RReadStream;
class RWriteStream;


///////////////////////////////////////////////////////////////////////////////
// TGameMessage
///////////////////////////////////////////////////////////////////////////////
// Basic inter-object message structure.
//
///////////////////////////////////////////////////////////////////////////////
class TGameMessage
	{
public:
	// Definitions.
	enum {             KMaxParamLen = 32 };
	typedef TUint16    TId;
	typedef TInt32     TResponse;
	typedef TInt32     TAddress;
public:
	// Address constants.
	const static TAddress KNullAddress;
	const static TAddress KBroadcastAddress;
public:
	// Construct and persist.
	TGameMessage();
	TGameMessage(TId aId, TAddress aSender, TAddress aReceiver);
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;
public:
	// Logging.
	void LogText(TDes&) const;
public:
	// Routing data.
	TId                 iId;
	TAddress            iSender;
	TAddress            iRecipient;
public:
	// User data.
	TInt32              iState;
	TBuf8<KMaxParamLen> iParams;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Message Responses.
///////////////////////////////////////////////////////////////////////////////
const TGameMessage::TResponse KGameMessageNotHandled = 0;
const TGameMessage::TResponse KGameOverResponse = KMaxTInt32;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Standard Message Ids.
///////////////////////////////////////////////////////////////////////////////
const TGameMessage::TId KMsgId_Null = 0x0000;
const TGameMessage::TId KMsgId_Base = 0x1000;

const TGameMessage::TId KMsgId_ObjectEnterState = KMsgId_Base + 1;
const TGameMessage::TId KMsgId_ObjectUpdate = KMsgId_Base + 2;
const TGameMessage::TId KMsgId_ObjectExitState = KMsgId_Base + 3;
const TGameMessage::TId KMsgId_ObjectThrust = KMsgId_Base + 4;
const TGameMessage::TId KMsgId_ObjectRotate = KMsgId_Base + 5;
const TGameMessage::TId KMsgId_ObjectShoot = KMsgId_Base + 6;
const TGameMessage::TId KMsgId_ObjectScorePoints = KMsgId_Base + 7;
const TGameMessage::TId KMsgId_ObjectStop = KMsgId_Base + 8;
const TGameMessage::TId KMsgId_ObjectCollision = KMsgId_Base + 9;
const TGameMessage::TId KMsgId_ObjectDie = KMsgId_Base + 10;
const TGameMessage::TId KMsgId_ObjectIncSpriteFrame = KMsgId_Base + 11;

const TGameMessage::TId KMsgId_ExplosionBase = KMsgId_Base + 100;
const TGameMessage::TId KMsgId_ExplosionAddRadius = KMsgId_Base + 101;

const TGameMessage::TId KMsgId_PlayerBase = KMsgId_Base + 200;
const TGameMessage::TId KMsgId_PlayerPowerupReload = KMsgId_PlayerBase + 1;
const TGameMessage::TId KMsgId_PlayerPowerupReloadExpiry = KMsgId_PlayerBase + 2;
const TGameMessage::TId KMsgId_PlayerPowerupShots = KMsgId_PlayerBase + 3;
const TGameMessage::TId KMsgId_PlayerPowerupShotsExpiry = KMsgId_PlayerBase + 4;
const TGameMessage::TId KMsgId_PlayerPowerupShield = KMsgId_PlayerBase + 5;
const TGameMessage::TId KMsgId_PlayerPowerupShieldExpiry = KMsgId_PlayerBase + 6;
const TGameMessage::TId KMsgId_PlayerPowerupActivate = KMsgId_PlayerBase + 20;
const TGameMessage::TId KMsgId_PlayerPowerupDeActivate = KMsgId_PlayerBase + 21;

const TGameMessage::TId KMsgId_UserBase = 0xf000;

///////////////////////////////////////////////////////////////////////////////
// Message Parameters
///////////////////////////////////////////////////////////////////////////////
// KMsgId_ObjectRotate
typedef TPckgC<TInt8> KMsgParamC_ObjectRotate; // TInt8 is rotate delta.
typedef TPckgBuf<TInt8> KMsgParamBuf_ObjectRotate; // TInt8 is rotate delta.
///////////////////////////////////////////////////////////////////////////////
// KMsgId_ObjectScorePoints
class TMsgScorePointsParam
	{
public:
	TInt16 iAllegience; // Allegience of scoring object.
	TInt32 iPoints;     // Points scored.
	};
typedef TPckgC<TMsgScorePointsParam> KMsgParamC_ScorePoints;
typedef TPckgBuf<TMsgScorePointsParam> KMsgParamBuf_ScorePoints;
///////////////////////////////////////////////////////////////////////////////
// KMsgId_PlayerPowerup....
class TMsgPowerupParam
	{
public:
	TUint16 iOnExpire; // Message to send on expiry.
	TUint16 iToExpire; // Message that causes expiry.
	TUint16 iDuration; // Num messages of iToExpire type received before expiry.
	TUint16 iValue;    // The effect.
	};
typedef TPckg<TMsgPowerupParam> KMsgParam_Powerup;
typedef TPckgC<TMsgPowerupParam> KMsgParamC_Powerup;
typedef TPckgBuf<TMsgPowerupParam> KMsgParamBuf_Powerup;
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// MGameMessageHandler
///////////////////////////////////////////////////////////////////////////////
// Message handler interface. Each game object is one of these.
//
///////////////////////////////////////////////////////////////////////////////
class MGameMessageHandler
	{
public:
	virtual ~MGameMessageHandler() {}
	virtual TGameMessage::TAddress& MessageAddress() = 0;
	virtual TGameMessage::TResponse OfferMessageL(const TGameMessage&) = 0;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// MGameMessageRouter
///////////////////////////////////////////////////////////////////////////////
// Object responsible for delivering messages to the correct object at the
// correct time.
//
///////////////////////////////////////////////////////////////////////////////
class MGameMessageRouter
	{
public:
	virtual ~MGameMessageRouter()   {}
	virtual void                    RegisterHandlerL(MGameMessageHandler&) = 0;
	virtual void                    UnRegisterHandler(MGameMessageHandler&) = 0;
	virtual TGameMessage::TResponse SendMessage(const TGameMessage&) = 0;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// RGameMessageForwarder
///////////////////////////////////////////////////////////////////////////////
class RGameMessageForwarder : public MGameMessageHandler, public MGameMessageRouter
	{
public:
	RGameMessageForwarder();
	~RGameMessageForwarder();
	void OpenL(MGameMessageHandler& aHandler, MGameMessageRouter& aRouter);
	void Close();
	virtual TGameMessage::TAddress& MessageAddress();
	virtual TGameMessage::TResponse OfferMessageL(const TGameMessage& aMessage);
	virtual void RegisterHandlerL(MGameMessageHandler& aHandler);
	virtual void UnRegisterHandler(MGameMessageHandler& aHandler);
	virtual TGameMessage::TResponse SendMessage(const TGameMessage& aMessage);
protected:
	TGameMessage::TAddress iAddress;
	MGameMessageHandler* iHandler;
	MGameMessageRouter* iRouter;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// RGameMessageBroadcaster
///////////////////////////////////////////////////////////////////////////////
class RGameMessageBroadcaster : public RGameMessageForwarder
	{
public:
	RGameMessageBroadcaster();
	~RGameMessageBroadcaster();
	inline void OpenL(MGameMessageHandler& aHandler, MGameMessageRouter& aRouter);
	inline void Close();
	virtual TGameMessage::TResponse OfferMessageL(const TGameMessage& aMessage);
	virtual TGameMessage::TResponse SendMessage(const TGameMessage& aMessage);
	};

///////////////////////////////////////////////////////////////////////////////
inline void RGameMessageBroadcaster::OpenL(MGameMessageHandler& aHandler, MGameMessageRouter& aRouter)
	{	RGameMessageForwarder::OpenL(aHandler, aRouter); }
inline void RGameMessageBroadcaster::Close()
	{	RGameMessageForwarder::Close(); }

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// TMessageCounter
///////////////////////////////////////////////////////////////////////////////
// When used with CMessageCounterList, allows a specific message to be sent
// when a specified message has been received a certain number of times.
//
///////////////////////////////////////////////////////////////////////////////
class TMessageCounter
	{
public:
	inline TMessageCounter() : iId(KMsgId_Null), iCount(0U) {}
public:
	TGameMessage::TId iId;
	TUint16 iCount;
	TGameMessage iExpiryMessage;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CMessageCounterList
///////////////////////////////////////////////////////////////////////////////
// Manages message counters for a single handler. All messages received by the
// handler should be passed to MessageL.
//
///////////////////////////////////////////////////////////////////////////////
class CMessageCounterList : public CBase
	{
public:
	CMessageCounterList();
	virtual ~CMessageCounterList();
	TInt AddCounter(const TMessageCounter&);
	void MessageL(MGameMessageRouter&, const TGameMessage&);
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;
	inline RArray<TMessageCounter>& Counters() { return iCounters; }
private:
	RArray<TMessageCounter> iCounters;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
