///////////////////////////////////////////////////////////////////////////////
//
// ObjectEmitter.cpp
//
//
// 
//
//
///////////////////////////////////////////////////////////////////////////////
#include "ObjectEmitter.h"
#include "ObjectController.h"
#include "GameResourceUtils.h"
#include <s32strm.h>
#include <barsread.h>

///////////////////////////////////////////////////////////////////////////////
// TGameObjectEmitter
///////////////////////////////////////////////////////////////////////////////
TGameObjectEmitter::TGameObjectEmitter()
: iSprite((TSpriteId)-1),
	iState(KStateIdNull),
	iAllegience(KAllegienceParent)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::EmitL(CGameObjectController& aController, const TMotionPhysics& aMotion)
	{
	CGameObject::TTemplate t(aController);
	t.iFacing = aMotion.Velocity().Direction();
	EmitObjectL(t, aMotion);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::InitTemplateL(CGameObject::TTemplate& aTemplate)
	{
	aTemplate.iSprite = iSprite;
	aTemplate.iState = iState;
	aTemplate.iAllegience = iAllegience;
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::EmitObjectL(CGameObject::TTemplate& aTemplate, const TMotionPhysics& aMotion)
	{
	aTemplate.iMotion = aMotion;
	EmitObjectL(aTemplate);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::EmitObjectL(CGameObject::TTemplate& aTemplate)
	{
	InitTemplateL(aTemplate);
	CGameObject* newObject = CGameObject::NewLC(aTemplate);
	aTemplate.iController.AddObjectL(newObject);
	CleanupStack::Pop(newObject);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::InternalizeL(TResourceReader& aReader)
	{
	iSprite = (TSpriteId)aReader.ReadInt32();
	iState = (TStateId)aReader.ReadInt32();
	iAllegience = TInt16(aReader.ReadInt16());
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::InternalizeL(RReadStream& aStream)
	{
	iSprite = (TSpriteId)aStream.ReadInt32L();
	iState = (TStateId)aStream.ReadInt32L();
	iAllegience = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void TGameObjectEmitter::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteInt32L(iSprite);
	aStream.WriteInt32L(iState);
	aStream.WriteInt16L(iAllegience);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// TGameFixedObjectEmitter
///////////////////////////////////////////////////////////////////////////////
TGameFixedObjectEmitter::TGameFixedObjectEmitter()
: iMass(1),
  iSpreadAngle(1)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::EmitChildL(const CGameObject& aEmitterObject)
	{
	iEmitterObject = &aEmitterObject;
	iFacing = aEmitterObject.Facing() + iFacingOffset;
	CGameObject::TTemplate t(aEmitterObject.Controller());
	EmitObjectL(t);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::EmitChildrenL(TInt aNumObjects, const CGameObject& aEmitterObject)
	{
	if(aNumObjects <= 0)
		{
		return;
		}
	iEmitterObject = &aEmitterObject;
	iFacing = aEmitterObject.Facing() + iFacingOffset;
	if(aNumObjects > 1)
		{
		TInt offset = iSpreadAngle.Value() * aNumObjects; // Total spread angle
		iFacing -= (offset >> 1);
		}
	CGameObject::TTemplate t(aEmitterObject.Controller());
	while(aNumObjects--)
		{
		EmitObjectL(t);
		iFacing += iSpreadAngle;
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::InitTemplateL(CGameObject::TTemplate& aTemplate)
	{
	ASSERT(iEmitterObject);
	aTemplate.iMotion.SetMass(iMass);
	aTemplate.iMotion.SetVelocity(iFacing.Vector(iSpeed));
	aTemplate.iFacing = iFacing;
	TGameObjectEmitter::InitTemplateL(aTemplate);
	CalculatePositionL(aTemplate);
	if(iAllegience == KAllegienceParent)
		{
		aTemplate.iAllegience = TInt16(iEmitterObject->Allegience());
		}
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::CalculatePositionL(CGameObject::TTemplate& aTemplate)
	{
	ASSERT(iEmitterObject);
	// Offset centre pos.
	const TTrigUnit radius(iEmitterObject->Radius());
	TVector v(aTemplate.iFacing.Vector(radius));
	TCoOrds pos(iEmitterObject->Centre());
	pos += v;
	// Convert from centre to TL.
	RGameSprite sprite(aTemplate.iController.Sprites());
	sprite.OpenL(aTemplate.iSprite);
	TSize size(sprite.FrameSize());
	sprite.Close();
	size.iWidth /= 2;
	size.iHeight /= 2;
	pos.iX -= size.iWidth;
	pos.iY -= size.iHeight;
	aTemplate.iMotion.SetPosition(pos);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::InternalizeL(TResourceReader& aReader)
	{
	TGameObjectEmitter::InternalizeL(aReader);
	iSpeed = GameResourceUtils::ReadTTrigUnit(aReader);
	iMass = GameResourceUtils::ReadTTrigUnit(aReader);
	iFacingOffset.InternalizeL(aReader);
	iSpreadAngle.InternalizeL(aReader);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::InternalizeL(RReadStream& aStream)
	{
	TGameObjectEmitter::InternalizeL(aStream);
	iSpeed.InternalizeL(aStream);
	iMass.InternalizeL(aStream);
	iFacingOffset.InternalizeL(aStream);
	iSpreadAngle.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void TGameFixedObjectEmitter::ExternalizeL(RWriteStream& aStream) const
	{
	TGameObjectEmitter::ExternalizeL(aStream);
	iSpeed.ExternalizeL(aStream);
	iMass.ExternalizeL(aStream);
	iFacingOffset.ExternalizeL(aStream);
	iSpreadAngle.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
