///////////////////////////////////////////////////////////////////////////////
//
// Objects.cpp
//
//
//
//
//
///////////////////////////////////////////////////////////////////////////////
#include "Game.pan"
#include "Objects.h"
#include "ObjectController.h"
#include "GameResourceUtils.h"
#include "GameTypes.hrh"
#include <s32strm.h>
#include <barsread.h>

const TUid KUidObjectCat       = { KGameObjType0_Object };
const TUid KUidObjectObject    = { KGameObjType1_Object };


///////////////////////////////////////////////////////////////////////////////
// CGameObject::TTemplate
//
//
///////////////////////////////////////////////////////////////////////////////
CGameObject::TTemplate::TTemplate(CGameObjectController& aController)
: iController(aController),
  iSprite(0),
	iThrustPower(1),
	iState(0),
	iAllegience(0)
	{
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::TTemplate::Set(const TTemplate& aTemplate)
	{
  iSprite = aTemplate.iSprite;
	iFacing = aTemplate.iFacing;
	iThrustPower = aTemplate.iThrustPower;
	iMotion = aTemplate.iMotion;
	iState = aTemplate.iState;
	iAllegience = aTemplate.iAllegience;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::TTemplate::InternalizeL(TResourceReader& aReader)
	{
	iSprite = aReader.ReadInt32();
	iFacing.InternalizeL(aReader);
	iThrustPower = GameResourceUtils::ReadTTrigUnit(aReader);
	iMotion.InternalizeL(aReader);
	iState = aReader.ReadInt32();
	iAllegience = TInt16(aReader.ReadInt16());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::TTemplate::InternalizeL(RReadStream& aStream)
	{
	iSprite = aStream.ReadInt32L();
	aStream >> iFacing;
	aStream >> iThrustPower;
	aStream >> iMotion;
	iState = aStream.ReadInt32L();
	iAllegience = aStream.ReadInt16L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::TTemplate::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteInt32L(iSprite);
	aStream << iFacing;
	aStream << iThrustPower;
	aStream << iMotion;
	aStream.WriteInt32L(iState);
	aStream.WriteInt16L(iAllegience);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// CGameObject::CStateData
///////////////////////////////////////////////////////////////////////////////
CGameObject::CStateData::CStateData()
: iCurrentState(KStateIdNull),
  iNextState(KStateIdNull)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject::CStateData::CStateData(CGameObjectState::TId aId)
: iCurrentState(aId),
  iNextState(KStateIdNull)
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject::CStateData::~CStateData()
	{
	delete iDataBuf;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::CStateData::InternalizeL(TResourceReader& aReader)
	{
	iCurrentState = aReader.ReadInt32();
	iNextState = aReader.ReadInt32();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::CStateData::InternalizeL(RReadStream& aStream)
	{
	iCurrentState = aStream.ReadInt32L();
	iNextState = aStream.ReadInt32L();
	iCounters.InternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::CStateData::ExternalizeL(RWriteStream& aStream) const
	{
	aStream.WriteInt32L(iCurrentState);
	aStream.WriteInt32L(iNextState);
	iCounters.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::CStateData::SetDataBufL(const TDesC8& aBuf)
	{
	HBufC8* buf = aBuf.AllocL();
	delete iDataBuf;
	iDataBuf = buf;
	}

///////////////////////////////////////////////////////////////////////////////
TPtr8 CGameObject::CStateData::DataBufL(TInt aLen)
	{
	if( (iDataBuf == NULL) || (iDataBuf->Length() < aLen) )
		{
		HBufC8* buf = HBufC8::NewL(Max(1, aLen));
		delete iDataBuf;
		iDataBuf = buf;
		}
	return iDataBuf->Des();
	}

///////////////////////////////////////////////////////////////////////////////
TPtrC8 CGameObject::CStateData::DataBuf() const
	{
	return (iDataBuf) ? TPtrC8(*iDataBuf) : TPtrC8(KNullDesC8);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameObject
//
//
///////////////////////////////////////////////////////////////////////////////
CGameObject::CGameObject(CGameObjectController& aController, CGameObjectState::TId aState)
: iController(aController),
  iSprite(aController.Sprites()),
	iStateData(aState),
  iMsgAddress(TGameMessage::KNullAddress),
  iObjectMotionBounds(&aController.ObjectMotionBounds())
	{
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject* CGameObject::NewLC(const CGameObject::TTemplate& aTemplate)
	{ // static
	CGameObject* o = new(ELeave) CGameObject(aTemplate.iController, aTemplate.iState);
	CleanupStack::PushL(o);
  o->iSprite.OpenL(aTemplate.iSprite);
  o->iFacing = aTemplate.iFacing;
  o->iThrustPower = aTemplate.iThrustPower;
  o->iMotion = aTemplate.iMotion;
  o->iAllegience = aTemplate.iAllegience;
	return o;
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject::~CGameObject()
	{
	iSprite.Close();
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TAddress& CGameObject::MessageAddress()
	{
	return iMsgAddress;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObject::OfferMessageL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = KGameMessageNotHandled;
	if(iMsgAddress != TGameMessage::KNullAddress)
		{
		CGameObjectState* state = iController.StateL(iStateData.CurrentState());
		if(state)
			{
			ret = state->HandleMessageL(aMessage);
			}
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObject::Type() const
	{
	const TUid uid = { iMsgAddress };
	return TUidType(KUidObjectCat, KUidObjectObject, uid);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::InternalizeL(TResourceReader& aReader)
	{
	iSprite.InternalizeL(aReader);
	iFacing.InternalizeL(aReader);
	iThrustPower = GameResourceUtils::ReadTTrigUnit(aReader);
	iMotion.InternalizeL(aReader);
	iStateData.InternalizeL(aReader);
	iAllegience = (TInt16)aReader.ReadInt16();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::InternalizeL(RReadStream& aStream)
	{
	aStream >> iSprite;
	aStream >> iFacing;
	aStream >> iThrustPower;
	aStream >> iMotion;
	aStream >> iStateData;
	iAllegience = (TInt16)aStream.ReadInt16L();
	iMsgAddress = aStream.ReadInt32L();
	iPoints = aStream.ReadInt32L();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::ExternalizeL(RWriteStream& aStream) const
	{
	aStream << iSprite;
	aStream << iFacing;
	aStream << iThrustPower;
	aStream << iMotion;
	aStream << iStateData;
	aStream.WriteInt16L(iAllegience);
	aStream.WriteInt32L(iMsgAddress);
	aStream.WriteInt32L(iPoints);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetId(TInt32 aId)
	{
	iMsgAddress = TGameMessage::TAddress(aId);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::Draw(CBitmapContext& aGc)
	{
	iSprite.DrawFrame(aGc, Position(), iObjectMotionBounds);
	}

///////////////////////////////////////////////////////////////////////////////
CGameObjectController& CGameObject::Controller() const
	{
	return iController;
	}

///////////////////////////////////////////////////////////////////////////////
const RGameSprite& CGameObject::Sprite() const
	{
	return iSprite;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::ChangeSpriteL(TSpriteId aId)
	{
	iSprite.OpenL(aId);
	}

///////////////////////////////////////////////////////////////////////////////
TTrigUnit CGameObject::Radius() const
	{
	const TSize size(Size());
	TTrigUnit t(Max(size.iWidth, size.iHeight));
	t /= 2;
	return t;
	}

///////////////////////////////////////////////////////////////////////////////
TSize CGameObject::Size() const
	{
	return iSprite.FrameSize();
	}

///////////////////////////////////////////////////////////////////////////////
TRect CGameObject::Rect() const
	{
	return TRect(TPoint(Position()), Size());
	}

///////////////////////////////////////////////////////////////////////////////
TCoOrds CGameObject::Centre() const
	{
	TTrigXY size(Size());
	size /= TTrigUnit(2);
	return size + Position();
	}

///////////////////////////////////////////////////////////////////////////////
const TCoOrds& CGameObject::Position() const
	{
	return iMotion.Position();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetPosition(const TCoOrds& aPos)
	{
	iMotion.SetPosition(aPos);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::UpdatePosition()
	{
	const TSize size(Size());
	iMotion.Update(&size, iObjectMotionBounds);
	}

///////////////////////////////////////////////////////////////////////////////
const TDirection& CGameObject::Facing() const
	{
	return iFacing;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetFacing(const TDirection& aDir)
	{
	iFacing = aDir;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::ChangeFacing(TInt aDelta)
	{
	iFacingDelta = TDirection(aDelta);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::UpdateFacing()
	{
	iFacing += iFacingDelta;
	ChangeFacing(0);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetMotionBounds(MMotionBounds& aBounds)
	{
	iObjectMotionBounds = &aBounds;
	}

///////////////////////////////////////////////////////////////////////////////
const TMotionPhysics& CGameObject::Motion() const
	{
	return iMotion;
	}

///////////////////////////////////////////////////////////////////////////////
TVector CGameObject::Velocity() const
	{
	return iMotion.Velocity();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetVelocity(const TVector& aVelocity)
	{
	iMotion.SetVelocity(aVelocity);
	}

///////////////////////////////////////////////////////////////////////////////
const TTrigUnit& CGameObject::ThrustPower() const
	{
	return iThrustPower;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetThrustPower(TTrigUnit aThrust)
	{
	iThrustPower = aThrust;
	}

///////////////////////////////////////////////////////////////////////////////
TTrigUnit CGameObject::Mass() const
	{
	return iMotion.Mass();
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetMass(TTrigUnit aMass)
	{
	iMotion.SetMass(aMass);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::ApplyForce(const TVector& aForce)
	{
	iMotion.ApplyForce(aForce);
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::MakeStationary()
	{
	iMotion.MakeStationary();
	}

///////////////////////////////////////////////////////////////////////////////
CGameObject::CStateData& CGameObject::StateData()
	{
	return iStateData;
	}

///////////////////////////////////////////////////////////////////////////////
const CGameObject::CStateData& CGameObject::StateData() const
	{
	return iStateData;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetAllegience(TInt aAllegience)
	{
	iAllegience = TInt16(aAllegience);
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameObject::Allegience() const
	{
	return iAllegience;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObject::SetPoints(TInt aPoints)
	{
	iPoints = aPoints;
	}

///////////////////////////////////////////////////////////////////////////////
TInt CGameObject::Points() const
	{
	return iPoints;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObject::SendMessage(TGameMessage::TId aId, TGameMessage::TAddress aTarget)
	{
	const TGameMessage msg(aId, iMsgAddress, aTarget);
	return SendMessage(msg);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObject::SendMessage(const TGameMessage& aMessage)
	{
	__ASSERT_DEBUG(iMsgAddress != TGameMessage::KNullAddress, Panic(EPanicObjectNotRegistered));
	__ASSERT_DEBUG(aMessage.iRecipient != TGameMessage::KNullAddress, Panic(EPanicObjectNotRegistered));
	return iController.SendMessage(aMessage);
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

