///////////////////////////////////////////////////////////////////////////////
//
// Objects.h
//
//
// Basic game object.
//
//
///////////////////////////////////////////////////////////////////////////////
#ifndef OBJECTS_H
#define OBJECTS_H

#include "GameContainer.h"
#include "Messages.h"
#include "State.h"
#include "Motion.h"
#include "Sprites.h"

const TInt KAllegienceParent = -1;
const TInt KAllegienceNone = 0;

class CGameObjectController;
///////////////////////////////////////////////////////////////////////////////
// CGameObject
///////////////////////////////////////////////////////////////////////////////
// Base class for all game objects. Although this class is a concrete class
// and defines a lot of data for a base class, it is largely data driven.
// Initialisation is intended to be (usually) from a file.
//
///////////////////////////////////////////////////////////////////////////////
// CGameObject::TTemplate
///////////////////////////////////////////////////////////////////////////////
// Provides the neccessary parameters to construct and perform a basic
// initialisation of a CGameObject.
//
///////////////////////////////////////////////////////////////////////////////
// CGameObject::TStateData
///////////////////////////////////////////////////////////////////////////////
// Never updated by CGameObject, this member is entirely controlled by the
// current CGameObjectState object associated with the iCurrentState member.
// The state objects will initialise it in OnEnterL(), update it in OnUpdateL()
// and reset it in OnExitL().
//
///////////////////////////////////////////////////////////////////////////////
class CGameObject
: public CBase,
  public MGameMessageHandler,
  public MGameStreamable,
  public MMotionBoid
	{
public:
	// CGameObject::TTemplate
	class TTemplate
		{
	public:
		TTemplate(CGameObjectController& aController);
		void Set(const TTemplate&);
		void InternalizeL(TResourceReader&);
		void InternalizeL(RReadStream&);
		void ExternalizeL(RWriteStream&) const;
	public:
		CGameObjectController& iController;
		TSpriteId iSprite;
		TDirection iFacing;
		TTrigUnit iThrustPower;
		TMotionPhysics iMotion;
		CGameObjectState::TId iState;
		TInt16 iAllegience;
		};
	// CGameObject::CStateData
	class CStateData : public CBase
		{
	public:
		CStateData();
		CStateData(CGameObjectState::TId);
		virtual ~CStateData();
	public:
		void InternalizeL(TResourceReader&);
		void InternalizeL(RReadStream&);
		void ExternalizeL(RWriteStream&) const;
	public:
		// State change.
		inline CGameObjectState::TId CurrentState() const           { return iCurrentState; }
		inline TBool StateChangeRequested() const                   { return (iNextState != KStateIdNull); }
		inline void ProcessStateChange()                            { iCurrentState = iNextState; iNextState = KStateIdNull; }
		inline void RequestStateChange(CGameObjectState::TId aNext) { iNextState = aNext; }
		// Message counters.
		inline RArray<TMessageCounter>& Counters()                                  { return iCounters.Counters(); }
		inline TInt AddCounter(const TMessageCounter& aCounter)                     { return iCounters.AddCounter(aCounter); }
		inline void MessageL(MGameMessageRouter& aRouter, const TGameMessage& aMsg) { iCounters.MessageL(aRouter, aMsg); }
		// Data.
		void SetDataBufL(const TDesC8&);
		TPtr8 DataBufL(TInt aLen);
		TPtrC8 DataBuf() const;
	private:
		CGameObjectState::TId iCurrentState;
		CGameObjectState::TId iNextState;
		CMessageCounterList iCounters;
		HBufC8* iDataBuf;
		};
	//
	// CGameObject
	//
public:
	// Construct.
	CGameObject(CGameObjectController& aController, CGameObjectState::TId aState=KStateIdNull);
	static CGameObject* NewLC(const TTemplate& aTemplate);
	// Destruct.
	virtual ~CGameObject();
	// Message Handling. (MGameMessageHandler)
  virtual TGameMessage::TAddress& MessageAddress();
	virtual TGameMessage::TResponse OfferMessageL(const TGameMessage&);
	// Persistence. (MGameStreamable)
	virtual TUidType Type() const;
	virtual void InternalizeL(TResourceReader&);
	virtual void InternalizeL(RReadStream&);
	virtual void ExternalizeL(RWriteStream&) const;
	virtual void SetId(TInt32);
	// Virtual Draw.
	virtual void Draw(CBitmapContext&);
public:
	// Controller.
	CGameObjectController& Controller() const;
	// Sprite.
	const RGameSprite& Sprite() const;
	void ChangeSpriteL(TSpriteId);
	inline void IncSpriteFrame() { iSprite.SetSpriteFrame(iSprite.Frame() + 1); }
	inline void SetSpriteFrame(TInt aFrame) { iSprite.SetSpriteFrame(aFrame); }
	inline void SetSpriteData(CGameSpriteData* aSpriteData, TBool aOwned) { iSprite.SetSpriteData(aSpriteData, aOwned); }
	// Position.
	virtual TCoOrds Centre() const;  // MMotionBoid
	const TCoOrds& Position() const; // N.B. Top left position, not centre.
	void SetPosition(const TCoOrds&);
	void UpdatePosition(); // According to motion.
	// Dimensions.
	virtual TTrigUnit Radius() const; // MMotionBoid
	TSize Size() const;
	TRect Rect() const;
	// Facing.
	const TDirection& Facing() const;
	void SetFacing(const TDirection&);
	void ChangeFacing(TInt aDelta); // Delayed until UpdateFacing() called.
	void UpdateFacing();
	// Motion.
	void SetMotionBounds(MMotionBounds&);
	const TMotionPhysics& Motion() const;
	virtual TVector Velocity() const; // MMotionBoid
	void SetVelocity(const TVector&);
	const TTrigUnit& ThrustPower() const;
	void SetThrustPower(TTrigUnit);
	TTrigUnit Mass() const;
	void SetMass(TTrigUnit);
	void ApplyForce(const TVector&);
	void MakeStationary();
	// State.
	CStateData& StateData();
	const CStateData& StateData() const;
	// Collisions.
	void SetAllegience(TInt);
	TInt Allegience() const;
	// Points.
	void SetPoints(TInt);
	TInt Points() const;
protected:
	// Message Sending.
	TGameMessage::TResponse SendMessage(TGameMessage::TId aId, TGameMessage::TAddress aTarget);
	TGameMessage::TResponse SendMessage(const TGameMessage&);

protected:
	mutable CGameObjectController& iController;
	RGameSprite iSprite;
	TDirection iFacing;
	TDirection iFacingDelta;
	TTrigUnit iThrustPower;
	TMotionPhysics iMotion;
	CStateData iStateData;
	TInt16 iAllegience;
	TGameMessage::TAddress iMsgAddress;
	TInt32 iPoints;
	MMotionBounds* iObjectMotionBounds;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
