///////////////////////////////////////////////////////////////////////////////
//
// PlayerStates.cpp
//
//
///////////////////////////////////////////////////////////////////////////////
#include "PlayerStates.h"
#include "ObjectController.h"
#include "GameResourceUtils.h"
#include "GameTypes.hrh"
#include <s32strm.h>
#include <barsread.h>

const TUid KUid0ObjectState = { KGameObjType0_ObjectState };

///////////////////////////////////////////////////////////////////////////////
// CGameObjectPlayerState
///////////////////////////////////////////////////////////////////////////////
CGameObjectPlayerState::CGameObjectPlayerState(CGameObjectController& aController)
: CGameObjectMotiveSolidState(aController)
	{
	iId = KGameObjType1_ObjectPlayerState;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectPlayerState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectPlayerState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectMotiveSolidState::InternalizeL(aReader);
	iWeapon.InternalizeL(aReader);
	iExhaust.InternalizeL(aReader);
	iNumShots = TUint8(aReader.ReadUint8());
	iReloadTime = TUint8(aReader.ReadUint8());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectMotiveSolidState::InternalizeL(aStream);
	iWeapon.InternalizeL(aStream);
	iExhaust.InternalizeL(aStream);
	iNumShots = TUint8(aStream.ReadUint8L());
	iReloadTime = TUint8(aStream.ReadUint8L());
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectMotiveSolidState::ExternalizeL(aStream);
	iWeapon.ExternalizeL(aStream);
	iExhaust.ExternalizeL(aStream);
	aStream.WriteUint8L(iNumShots);
	aStream.WriteUint8L(iReloadTime);
	}

///////////////////////////////////////////////////////////////////////////////
TBool CGameObjectPlayerState::TestCollision(const CGameObject& aObject, TInt aAllegience)
	{
	TBool ret = CGameObjectState::TestCollision(aObject, aAllegience);
	if(!ret)
		{
		// Can also collide with pickups.
		CGameObjectState* state = iController.StateL(aObject.StateData().CurrentState());
		if(state)
			ret = (state->Type()[1].iUid == KGameObjType1_ObjectPickupState);
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::OnEnterL(const TGameMessage& aMessage)
	{
	(void)CGameObjectMotiveSolidState::OnEnterL(aMessage);
	TStateDataBuf buf;
	buf.iReloadStatus = 0;
	buf.iReloadTime = iReloadTime;
	buf.iNumShots = iNumShots;
	SetObjectStateDataBufL(ObjectL(aMessage), buf);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::OnUpdateL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse ret = CGameObjectMotiveSolidState::OnUpdateL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	// Update sprite frame for facing.
	o.SetSpriteFrame(o.Facing().Value());
	// Update reload timer.
	TStateDataBuf buf;
	ObjectStateDataBuf(o, buf);
	if(buf.iReloadStatus)
		{
		--buf.iReloadStatus;
		SetObjectStateDataBufL(o, buf);
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::OnMessageL(const TGameMessage& aMessage)
	{
	TGameMessage::TResponse	ret = CGameObjectMotiveSolidState::OnMessageL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	switch(aMessage.iId)
		{
	case KMsgId_ObjectThrust:
		ret = HandleThrustL(o, aMessage);
		break;
	case KMsgId_ObjectStop:
		o.MakeStationary();
		ret = 1;
		break;
	case KMsgId_ObjectRotate:
		ret = HandleRotate(o, aMessage);
		break;
	case KMsgId_ObjectShoot:
		ret = HandleShootL(o, aMessage);
		break;
	case KMsgId_ObjectScorePoints:
		ret = HandleScorePoints(o, aMessage);
		break;
	case KMsgId_PlayerPowerupReload:
	case KMsgId_PlayerPowerupShots:
	case KMsgId_PlayerPowerupShield:
		ret = HandlePowerupL(o, aMessage);
		break;
	case KMsgId_PlayerPowerupReloadExpiry:
	case KMsgId_PlayerPowerupShotsExpiry:
	case KMsgId_PlayerPowerupShieldExpiry:
		ret = HandlePowerupExpiryL(o, aMessage);
		break;
	case KMsgId_PlayerPowerupActivate:
	case KMsgId_PlayerPowerupDeActivate:
		ret = ActivatePowerupL(o, aMessage, aMessage.iId == KMsgId_PlayerPowerupActivate);
		break;
	default:
		break;
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::OnCollisionL(const TGameMessage& aMessage)
	{
	return CGameObjectActiveState::OnCollisionL(aMessage);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandleThrustL(CGameObject& aObject, const TGameMessage&)
	{
	const TTrigUnit& thrust = aObject.ThrustPower();
	const TDirection& facing = aObject.Facing();
	aObject.ApplyForce( facing.Vector(thrust) );
	iExhaust.EmitChildL(aObject);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandleRotate(CGameObject& aObject, const TGameMessage& aMessage)
	{
	KMsgParamBuf_ObjectRotate param;
	const TDesC8& msgParams = aMessage.iParams;
	param.Copy(msgParams.Ptr(), param.MaxSize());
	aObject.ChangeFacing(param());
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandleShootL(CGameObject& aObject, const TGameMessage&)
	{
	TStateDataBuf buf;
	ObjectStateDataBuf(aObject, buf);
	if(buf.iReloadStatus == 0)
		{
		iWeapon.EmitChildrenL(buf.iNumShots, aObject);
		aObject.SetPoints(aObject.Points() - 1);
		buf.iReloadStatus = buf.iReloadTime;
		SetObjectStateDataBufL(aObject, buf);
		return 1;
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandleScorePoints(CGameObject& aObject, const TGameMessage& aMessage)
	{
	KMsgParamBuf_ScorePoints param;
	const TDesC8& msgParams = aMessage.iParams;
	param.Copy(msgParams.Ptr(), param.MaxSize());
	if(param().iAllegience == aObject.Allegience())
		{
		aObject.SetPoints(param().iPoints + aObject.Points());
		return 1;
		}
	return KGameMessageNotHandled;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::ActivatePowerupL(CGameObject& aObject, const TGameMessage&, TBool aActive)
	{
	// TODO: Currently this assumes shield is selected.

	TGameMessage::TResponse ret = KGameMessageNotHandled;
	// EPlayerKeyActivate key event received.
	TStateDataBuf buf;
	ObjectStateDataBuf(aObject, buf);
	TGameMessage msg(KMsgId_Null, aObject.MessageAddress(), aObject.MessageAddress());
	if(!buf.IsShieldUp() && aActive)
		{
		// Activate in response to user key press.
		msg.iId = KMsgId_PlayerPowerupShield;
		TMsgPowerupParam dummy;
		KMsgParam_Powerup paramsPtr(dummy);
		paramsPtr.Set((TUint8*)msg.iParams.Ptr(), sizeof(TMsgPowerupParam), sizeof(TMsgPowerupParam));
		paramsPtr().iOnExpire = KMsgId_PlayerPowerupDeActivate;
		paramsPtr().iToExpire = KMsgId_ObjectUpdate;
		if(!buf.iShieldDuration)
			{
			buf.iShieldDuration = iShieldDuration; // Initialising first time only.
			SetObjectStateDataBufL(aObject, buf);
			}
		paramsPtr().iDuration = buf.iShieldDuration;
		paramsPtr().iValue = 0;
		ret = HandlePowerupL(aObject, msg);
		}
	else if(buf.IsShieldUp() && !aActive)
		{
		// Deactivate in response to user key release.
		msg.iId = KMsgId_PlayerPowerupShieldExpiry;
		buf.LowerShield();
		// N.B. Don't reset the shield duration here, user can activate again by pressing key.
		SetObjectStateDataBufL(aObject, buf);
		ret = 1;
		}
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandlePowerupL(CGameObject& aObject, const TGameMessage& aMessage)
	{
	// Reload time, num shots, need to go in state data.
	// Change value in object state data, and add a timeout.

	TStateDataBuf buf;
	ObjectStateDataBuf(aObject, buf);
	KMsgParamBuf_Powerup paramBuf;
	paramBuf.Copy(aMessage.iParams);
	const TMsgPowerupParam& param = paramBuf();
	switch(aMessage.iId)
		{
	case KMsgId_PlayerPowerupReload:
		buf.iReloadTime = TUint8(Max(0, TInt(buf.iReloadTime) - param.iValue));
		break;
	case KMsgId_PlayerPowerupShots:
		buf.iNumShots = TUint8(param.iValue);
		break;
	case KMsgId_PlayerPowerupShield: // In response to user key event.
		buf.RaiseShield();
		break;
	default:
		return KGameMessageNotHandled;
		}
	if( (param.iOnExpire != KMsgId_Null) && (param.iToExpire != KMsgId_Null) && (param.iDuration > 0) )
		{
		// Check if pickup already exists and extend duration.
		RArray<TMessageCounter>& msgCounters = aObject.StateData().Counters();
		TInt c = msgCounters.Count();
		while(c--)
			{
			TMessageCounter& counter = msgCounters[c];
			if(counter.iExpiryMessage.iId == param.iOnExpire)
				{
				// Same expiry message, so same pickup. Extend duration.
				counter.iCount = TUint16(counter.iCount + param.iDuration);
				break;
				}
			}
		if(c < 0)
			{
			// Add new counter.
			const TGameMessage::TAddress address(aObject.MessageAddress());
			TGameMessage msg(param.iOnExpire, address, address);
			msg.iState = aMessage.iState;
			AddMessageCounterL(aObject, param.iToExpire, param.iDuration, msg);
			}
		}
	SetObjectStateDataBufL(aObject, buf);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerState::HandlePowerupExpiryL(CGameObject& aObject, const TGameMessage& aMessage)
	{
	TStateDataBuf buf;
	ObjectStateDataBuf(aObject, buf);
	switch(aMessage.iId)
		{
	case KMsgId_PlayerPowerupReloadExpiry:
		buf.iReloadTime = iReloadTime;
		break;
	case KMsgId_PlayerPowerupShotsExpiry:
		buf.iNumShots = iNumShots;
		break;
	case KMsgId_PlayerPowerupShieldExpiry: // In response to timer expiry, (not user key release).
		buf.LowerShield();
		buf.iShieldDuration = 0;
		break;
	default:
		return KGameMessageNotHandled;
		break;
		}
	SetObjectStateDataBufL(aObject, buf);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerState::ObjectStateDataBuf(const CGameObject& aObject, TStateDataBuf& aBuf) const
	{
	TPtrC8 des = aObject.StateData().DataBuf();
	if(des.Length() == sizeof(TStateDataBuf))
		{
		TPckg<TStateDataBuf> buf(aBuf);
		buf.Copy(des);
		}
	else
		{
		Mem::FillZ(&aBuf, sizeof(TStateDataBuf));
		}
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerState::SetObjectStateDataBufL(CGameObject& aObject, const TStateDataBuf& aBuf) const
	{
	aObject.StateData().SetDataBufL(TPckgC<TStateDataBuf>(aBuf));
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// CGameObjectPlayerPickupState
///////////////////////////////////////////////////////////////////////////////
CGameObjectPlayerPickupState::CGameObjectPlayerPickupState(CGameObjectController& aController)
: CGameObjectMotiveSolidState(aController)
	{
	iId = KGameObjType1_ObjectPickupState;
	}

///////////////////////////////////////////////////////////////////////////////
TUidType CGameObjectPlayerPickupState::Type() const
	{
	const TUid uid = { KGameObjType1_ObjectPickupState };
	return TUidType(KUid0ObjectState, uid, TUid::Uid(iId));
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerPickupState::InternalizeL(TResourceReader& aReader)
	{
	CGameObjectMotiveSolidState::InternalizeL(aReader);
	iMsgId = TGameMessage::TId(aReader.ReadUint16());
	TMsgPowerupParam msgParams;
	msgParams.iOnExpire = TUint16(aReader.ReadUint16());
	msgParams.iToExpire = TUint16(aReader.ReadUint16());
	msgParams.iDuration = TUint16(aReader.ReadUint16());
	msgParams.iValue = TUint16(aReader.ReadUint16());
	iMsgParams = KMsgParamC_Powerup(msgParams);
	iPoints = TInt16(aReader.ReadInt16());
	iSpeedRange = TUint16(aReader.ReadUint16());
	iSpeedDeviation = GameResourceUtils::ReadTTrigUnit(aReader);
	iEmitter.InternalizeL(aReader);
	iSpeed = iEmitter.iSpeed;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerPickupState::InternalizeL(RReadStream& aStream)
	{
	CGameObjectMotiveSolidState::InternalizeL(aStream);
	iMsgId = TGameMessage::TId(aStream.ReadUint16L());
	aStream >> iMsgParams;
	iPoints = aStream.ReadInt16L();
	iSpeedRange = aStream.ReadUint16L();
	iSpeedDeviation.InternalizeL(aStream);
	iEmitter.InternalizeL(aStream);
	iSpeed = iEmitter.iSpeed;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerPickupState::ExternalizeL(RWriteStream& aStream) const
	{
	CGameObjectMotiveSolidState::ExternalizeL(aStream);
	aStream.WriteUint16L(iMsgId);
	aStream << iMsgParams;
	aStream.WriteInt16L(iPoints);
	aStream.WriteUint16L(iSpeedRange);
	iSpeedDeviation.ExternalizeL(aStream);
	iEmitter.ExternalizeL(aStream);
	}

///////////////////////////////////////////////////////////////////////////////
TBool CGameObjectPlayerPickupState::TestCollision(const CGameObject& aObject, TInt)
	{
	// Can only collide with players.
	if(aObject.StateData().CurrentState() == KGameObjType1_ObjectPlayerState)
		{
		return ETrue;
		}
	return EFalse;
	}

///////////////////////////////////////////////////////////////////////////////
void CGameObjectPlayerPickupState::CreateNewL(const CGameObject& aParent)
	{
	iEmitter.iSpeed = iController.Random(iSpeed, TInt(iSpeedRange), iSpeedDeviation);
	iEmitter.iFacingOffset = TDirection(iController.Random(TDirection::KResolution));
	iEmitter.EmitChildL(aParent);
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerPickupState::OnEnterL(const TGameMessage& aMessage)
	{
	CGameObjectMotiveSolidState::OnEnterL(aMessage);
	CGameObject& o = ObjectL(aMessage);
	o.SetPoints(iPoints);
	return 1;
	}

///////////////////////////////////////////////////////////////////////////////
TGameMessage::TResponse CGameObjectPlayerPickupState::OnCollisionL(const TGameMessage& aMessage)
	{
	if(ObjectL(aMessage, CGameObjectState::ESender).StateData().CurrentState() != KGameObjType1_ObjectPlayerState)
		return KGameMessageNotHandled; // Ignore collisions with non-player objects.

	TGameMessage::TResponse ret = CGameObjectMotiveSolidState::OnCollisionL(aMessage);

	// Collided with player, send points scored and powerup messages.
	SendPointsScoreMessageL(aMessage);
	if( (iMsgId != KMsgId_Null) && (iMsgId != KMsgId_ObjectScorePoints) )
		{
		TGameMessage msg(iMsgId, aMessage.iRecipient, aMessage.iSender);
		msg.iParams = iMsgParams;
		msg.iState = iId;
		iController.SendMessage(msg);
		}
	ret = 1;
	return ret;
	}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

