///////////////////////////////////////////////////////////////////////////////
//
// Trigonometry.h
//
//
//
///////////////////////////////////////////////////////////////////////////////
#ifndef TRIGONOMETRY_H
#define TRIGONOMETRY_H

#include <e32std.h>
#include <s32std.h>
#include <e32math.h>
class TResourceReader;

///////////////////////////////////////////////////////////////////////////////
// TTrigUnit
///////////////////////////////////////////////////////////////////////////////
// Encapsulation of a floating point 'trigonometry unit'.
// This enables the current, (probably somewhat naive), TReal implementation
// to be replaced, e.g. by a fixed point implementation, in future.
///////////////////////////////////////////////////////////////////////////////
class TTrigUnit
	{
public:
	inline TTrigUnit();
	inline TTrigUnit(const TTrigUnit&);
	inline TTrigUnit(TInt aInt);
	inline TTrigUnit(TReal);
public:
	inline void InternalizeL(RReadStream&);
	inline void ExternalizeL(RWriteStream&) const;
	inline void Set(TInt aInt);
	inline void Set(TReal aReal);
	inline void Zero();
public:
	inline void Format(TDes& aDes) const;
	inline TInt NearestInt() const;
	inline TInt Int() const;
	inline TInt Fraction() const;
public:
	inline TTrigUnit Pow2() const;
	inline TTrigUnit Sqrt() const;
public:
	inline TTrigUnit& operator =(const TTrigUnit&);
	inline TTrigUnit& operator+=(const TTrigUnit&);
	inline TTrigUnit& operator-=(const TTrigUnit&);
	inline TTrigUnit& operator*=(const TTrigUnit&);
	inline TTrigUnit& operator/=(const TTrigUnit&);
	inline TTrigUnit& operator%=(const TTrigUnit&);
public:
	inline TTrigUnit operator+() const;
	inline TTrigUnit operator-() const;
public:
	inline TTrigUnit operator+(const TTrigUnit&) const;
	inline TTrigUnit operator-(const TTrigUnit&) const;
	inline TTrigUnit operator*(const TTrigUnit&) const;
	inline TTrigUnit operator/(const TTrigUnit&) const;
	inline TTrigUnit operator%(const TTrigUnit&) const;
public:
	inline TBool operator==(const TTrigUnit&) const;
	inline TBool operator!=(const TTrigUnit&) const;
	inline TBool operator<=(const TTrigUnit&) const;
	inline TBool operator>=(const TTrigUnit&) const;
	inline TBool operator< (const TTrigUnit&) const;
	inline TBool operator> (const TTrigUnit&) const;
private:
	TReal iValue;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// TTrigXY
///////////////////////////////////////////////////////////////////////////////
// A 2D point trig structure.
///////////////////////////////////////////////////////////////////////////////
class TTrigXY
	{
public:
	inline TTrigXY();
	inline TTrigXY(TTrigUnit aX, TTrigUnit aY);
	inline TTrigXY(const TSize& aSize);
	inline TTrigXY(const TPoint& aPoint);
	inline void Set(TTrigUnit aX, TTrigUnit aY);
	inline void Zero();
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;
public:
	inline TTrigXY& operator= (const TTrigXY&);
	inline TTrigXY& operator+=(const TTrigXY&);
	inline TTrigXY& operator-=(const TTrigXY&);
	inline TTrigXY& operator*=(const TTrigXY&);
	inline TTrigXY& operator/=(const TTrigXY&);
	inline TTrigXY& operator%=(const TTrigXY&);
	inline TTrigXY& operator= (TTrigUnit);
	inline TTrigXY& operator+=(TTrigUnit);
	inline TTrigXY& operator-=(TTrigUnit);
	inline TTrigXY& operator*=(TTrigUnit);
	inline TTrigXY& operator/=(TTrigUnit);
	inline TTrigXY& operator%=(TTrigUnit);
	inline TTrigXY operator+(const TTrigXY&) const;
	inline TTrigXY operator-(const TTrigXY&) const;
	inline TTrigXY operator*(const TTrigXY&) const;
	inline TTrigXY operator/(const TTrigXY&) const;
	inline TTrigXY operator%(const TTrigXY&) const;
	inline TTrigXY operator+(TTrigUnit) const;
	inline TTrigXY operator-(TTrigUnit) const;
	inline TTrigXY operator*(TTrigUnit) const;
	inline TTrigXY operator/(TTrigUnit) const;
	inline TTrigXY operator%(TTrigUnit) const;
	inline TBool   operator==(const TTrigXY&) const;
	inline TBool   operator!=(const TTrigXY&) const;
	operator TPoint() const { return TPoint(iX.NearestInt(), iY.NearestInt()); }
	operator TSize() const  { return TSize(iX.NearestInt(), iY.NearestInt()); }
public:
	TTrigUnit iX;
	TTrigUnit iY;
	};

///////////////////////////////////////////////////////////////////////////////
typedef TTrigXY TCoOrds;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// TDirection
///////////////////////////////////////////////////////////////////////////////
// A game direction, resolved to 16 units.
///////////////////////////////////////////////////////////////////////////////
class TDirection
	{
public:
	enum { KResolution = 16 };
public:
	inline TDirection();
	inline explicit TDirection(TInt aValue);
	explicit TDirection(TTrigUnit aAngle);
	void InternalizeL(TResourceReader&);
	void InternalizeL(RReadStream&);
	void ExternalizeL(RWriteStream&) const;
public:
	inline TInt Value() const;
	inline void Reverse();
	TDirection Opposite() const;
	TTrigXY Vector(const TTrigUnit& aMagnitude) const;
public:
	inline TBool operator==(const TDirection&) const;
	inline TBool operator!=(const TDirection&) const;
	TDirection& operator-=(TInt);
	TDirection& operator+=(TInt);
	TDirection operator-(TInt) const;
	TDirection operator+(TInt) const;
	inline TDirection& operator-=(const TDirection&);
	inline TDirection& operator+=(const TDirection&);
	inline TDirection operator-(const TDirection&) const;
	inline TDirection operator+(const TDirection&) const;
	inline TDirection& operator--(); // Prefix
	inline TDirection& operator++(); // Prefix
	inline const TDirection operator--(TInt); // Postfix
	inline const TDirection operator++(TInt); // Postfix
private:
	TInt iValue;
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Trigonometry
///////////////////////////////////////////////////////////////////////////////
// Static trig functions using TDirection.
///////////////////////////////////////////////////////////////////////////////
class Trigonometry
	{
public:
	static TTrigUnit Sin(TDirection);
	static TTrigUnit Cos(TDirection);
	static TDirection ASin(TTrigUnit);
	static TDirection ACos(TTrigUnit);
public:
	static TTrigUnit KPi();
	};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#include "Trigonometry.inl"

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#endif
